# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

import os
import os.path as osp
from typing import Callable, List, Optional
import numpy as np
import torch
from torch_geometric.data import (
    Data,
    InMemoryDataset,
    download_url,
    extract_tar,
    extract_zip,
)

# extracted from PyG MD17 dataset class


class MD17(InMemoryDataset):

    gdml_url = "http://quantum-machine.org/gdml/data/npz"
    revised_url = (
        "https://archive.materialscloud.org/record/"
        "file?filename=rmd17.tar.bz2&record_id=466"
    )

    file_names = {
        "benzene": "md17_benzene2017.npz",
        "uracil": "md17_uracil.npz",
        "naphtalene": "md17_naphthalene.npz",
        "aspirin": "md17_aspirin.npz",
        "salicylic_acid": "md17_salicylic.npz",
        "malonaldehyde": "md17_malonaldehyde.npz",
        "ethanol": "md17_ethanol.npz",
        "toluene": "md17_toluene.npz",
        "paracetamol": "paracetamol_dft.npz",
        "azobenzene": "azobenzene_dft.npz",
        "revised_benzene": "rmd17_benzene.npz",
        "revised_uracil": "rmd17_uracil.npz",
        "revised_naphthalene": "rmd17_naphthalene.npz",
        "revised_aspirin": "rmd17_aspirin.npz",
        "revised_salicylic_acid": "rmd17_salicylic.npz",
        "revised_malonaldehyde": "rmd17_malonaldehyde.npz",
        "revised_ethanol": "rmd17_ethanol.npz",
        "revised_toluene": "rmd17_toluene.npz",
        "revised_paracetamol": "rmd17_paracetamol.npz",
        "revised_azobenzene": "rmd17_azobenzene.npz",
        "benzene_CCSD_T": "benzene_ccsd_t.zip",
        "aspirin_CCSD": "aspirin_ccsd.zip",
        "malonaldehyde_CCSD_T": "malonaldehyde_ccsd_t.zip",
        "ethanol_CCSD_T": "ethanol_ccsd_t.zip",
        "toluene_CCSD_T": "toluene_ccsd_t.zip",
        "benzene_FHI-aims": "benzene2018_dft.npz",
    }

    def __init__(
        self,
        root: str,
        molecules: str,
        train: Optional[bool] = None,
        transform: Optional[Callable] = None,
        pre_transform: Optional[Callable] = None,
        pre_filter: Optional[Callable] = None,
    ):
        name = molecules
        if name not in self.file_names:
            raise ValueError(f"Unknown dataset name '{name}'")

        self.name = name
        self.revised = "revised" in name
        self.ccsd = "CCSD" in self.name

        super().__init__(root, transform, pre_transform, pre_filter)

        if len(self.processed_file_names) == 1 and train is not None:
            raise ValueError(
                f"'{self.name}' dataset does not provide pre-defined splits "
                f"but the 'train' argument is set to '{train}'"
            )
        elif len(self.processed_file_names) == 2 and train is None:
            raise ValueError(
                f"'{self.name}' dataset does provide pre-defined splits but "
                f"the 'train' argument was not specified"
            )

        idx = 0 if train is None or train else 1
        self.data, self.slices = torch.load(self.processed_paths[idx])

    def mean(self) -> float:
        return float(self._data.energy.mean())

    @property
    def raw_dir(self) -> str:
        if self.revised:
            return osp.join(self.root, "raw")
        return osp.join(self.root, self.name, "raw")

    @property
    def processed_dir(self) -> str:
        return osp.join(self.root, "processed", self.name)

    @property
    def raw_file_names(self) -> str:
        name = self.file_names[self.name]
        if self.revised:
            return osp.join("rmd17", "npz_data", name)
        elif self.ccsd:
            return name[:-4] + "-train.npz", name[:-4] + "-test.npz"
        return name

    @property
    def processed_file_names(self) -> List[str]:
        if self.ccsd:
            return ["train.pt", "test.pt"]
        else:
            return ["data.pt"]

    def download(self):
        if self.revised:
            path = download_url(self.revised_url, self.raw_dir)
            extract_tar(path, self.raw_dir, mode="r:bz2")
            os.unlink(path)
        else:
            url = f"{self.gdml_url}/{self.file_names[self.name]}"
            path = download_url(url, self.raw_dir)
            if self.ccsd:
                extract_zip(path, self.raw_dir)
                os.unlink(path)

    def process(self):
        it = zip(self.raw_paths, self.processed_paths)
        for raw_path, processed_path in it:
            raw_data = np.load(raw_path)

            if self.revised:
                z = torch.from_numpy(raw_data["nuclear_charges"]).long()
                pos = torch.from_numpy(raw_data["coords"]).float()
                energy = torch.from_numpy(raw_data["energies"]).float()
                force = torch.from_numpy(raw_data["forces"]).float()
            else:
                z = torch.from_numpy(raw_data["z"]).long()
                pos = torch.from_numpy(raw_data["R"]).float()
                energy = torch.from_numpy(raw_data["E"]).float()
                force = torch.from_numpy(raw_data["F"]).float()

            data_list = []
            for i in range(pos.size(0)):
                data = Data(z=z, pos=pos[i], y=energy[i].unsqueeze(-1), neg_dy=force[i])
                if self.pre_filter is not None and not self.pre_filter(data):
                    continue
                if self.pre_transform is not None:
                    data = self.pre_transform(data)
                data_list.append(data)

            torch.save(self.collate(data_list), processed_path)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({len(self)}, name='{self.name}')"



# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

import torch
from typing import Optional, Tuple
from torch import Tensor, nn
from torchmdnet.models.utils import (
    CosineCutoff,
    OptimizedDistance,
    rbf_class_mapping,
    act_class_mapping,
)

__all__ = ["TensorNet"]


def _decompose_tensor(tensor):
    """Full tensor decomposition into irreducible components.

    shapes are [N, 3, 3, F]
    """
    A = 0.5 * (tensor - tensor.transpose(1, 2))
    S = tensor - A
    I = (tensor.diagonal(dim1=1, dim2=2)).mean(-1)
    S = S - I_to_tensor(I)
    return I, A, S


def _compose_tensor(I, A, S):
    """Compose tensor from irreducible components.

    input shapes:
    I: [batch_size, hidden_channels]
    A: [batch_size, 3, 3, hidden_channels]
    S: [batch_size, 3, 3, hidden_channels]
    output shape:
    [batch_size, 3, 3, hidden_channels]"""
    return I_to_tensor(I) + A + S


def _tensor_matmul_o3(Y, msg):
    A = torch.matmul(msg.permute(0, 3, 1, 2), Y.permute(0, 3, 1, 2)).permute(0, 2, 3, 1)
    B = torch.matmul(Y.permute(0, 3, 1, 2), msg.permute(0, 3, 1, 2)).permute(0, 2, 3, 1)
    return A + B


def _tensor_matmul_so3(Y, msg):
    return torch.matmul(Y.permute(0, 3, 1, 2), msg.permute(0, 3, 1, 2)).permute(
        0, 2, 3, 1
    )


try:
    from torchmdnet.extensions.warp_ops import (
        graph_transform,
        fn_radial_message_passing as fn_tensornet_embedding_message_passing,
        fn_message_passing as fn_tensornet_interaction_message_passing,
        fn_compose_tensor as compose_tensor,
        fn_decompose_tensor as decompose_tensor,
        fn_tensor_matmul_o3_3x3 as tensor_matmul_o3,
        fn_tensor_matmul_so3_3x3 as tensor_matmul_so3,
        fn_tensor_norm3,
    )

    OPT = True
except ImportError:
    import warnings

    warnings.warn(
        "Could not import warp_ops (warp-lang not installed?). "
        "Falling back to pure-Python TensorNet ops. "
        "Install warp-lang for better performance: pip install warp-lang",
        ImportWarning,
        stacklevel=2,
    )
    OPT = False
    compose_tensor = _compose_tensor
    decompose_tensor = _decompose_tensor
    tensor_matmul_o3 = _tensor_matmul_o3
    tensor_matmul_so3 = _tensor_matmul_so3


def vector_to_skewtensor(vector):
    """Creates a skew-symmetric tensor from a vector."""
    B, _, F = vector.shape
    zero = torch.zeros((B, F), device=vector.device, dtype=vector.dtype)
    tensor = torch.stack(
        (
            zero,
            -vector[..., 2, :],
            vector[..., 1, :],
            vector[..., 2, :],
            zero,
            -vector[..., 0, :],
            -vector[..., 1, :],
            vector[..., 0, :],
            zero,
        ),
        dim=1,
    )
    tensor = tensor.view(B, 3, 3, F)
    return tensor


def skewtensor_to_vector(tensor):
    """Converts 3x3 skewtensor A to a vector

    input shape must be [N, 3, 3, F]
    output shape is [N, 3, F]
    """

    tensor = tensor.flatten(1, 2)
    vector = 0.5 * torch.stack(
        (
            tensor[:, 7, :] - tensor[:, 5, :],
            tensor[:, 2, :] - tensor[:, 6, :],
            tensor[:, 3, :] - tensor[:, 1, :],
        ),
        dim=1,
    )
    return vector


def I_to_tensor(I):
    """Converts scalar to 3x3 I tensor with shape [N, 3, 3, F]"""
    return (
        I[:, None, None, :]
        * torch.eye(3, 3, device=I.device, dtype=I.dtype)[None, ..., None]
    )


def outer_to_symtensor(tensor):
    """Creates a symmetric traceless tensor from the outer product tensor

    shapes are [N, 3, 3, F]
    """
    S = 0.5 * (tensor + tensor.transpose(1, 2))
    I = (tensor.diagonal(dim1=1, dim2=2)).mean(-1)
    S = S - I_to_tensor(I)
    return S


def tensor_norm(tensor):
    """Computes Frobenius norm."""
    return (tensor**2).sum((1, 2))


class TensorNet(nn.Module):
    r"""TensorNet's architecture. From
    TensorNet: Cartesian Tensor Representations for Efficient Learning of Molecular Potentials; G. Simeon and G. de Fabritiis.
    NeurIPS 2023.

    This function optionally supports periodic boundary conditions with arbitrary triclinic boxes.
    For a given cutoff, :math:`r_c`, the box vectors :math:`\vec{a},\vec{b},\vec{c}` must satisfy certain requirements:

    .. math::

      \begin{align*}
      a_y = a_z = b_z &= 0 \\
      a_x, b_y, c_z &\geq 2 r_c \\
      a_x &\geq 2  b_x \\
      a_x &\geq 2  c_x \\
      b_y &\geq 2  c_y
      \end{align*}

    These requirements correspond to a particular rotation of the system and reduced form of the vectors, as well as the requirement that the cutoff be no larger than half the box width.

    Args:
        hidden_channels (int, optional): Hidden embedding size.
            (default: :obj:`128`)
        num_layers (int, optional): The number of interaction layers.
            (default: :obj:`2`)
        num_rbf (int, optional): The number of radial basis functions :math:`\mu`.
            (default: :obj:`32`)
        rbf_type (string, optional): The type of radial basis function to use.
            (default: :obj:`"expnorm"`)
        trainable_rbf (bool, optional): Whether to train RBF parameters with
            backpropagation. (default: :obj:`False`)
        activation (string, optional): The type of activation function to use.
            (default: :obj:`"silu"`)
        cutoff_lower (float, optional): Lower cutoff distance for interatomic interactions.
            (default: :obj:`0.0`)
        cutoff_upper (float, optional): Upper cutoff distance for interatomic interactions.
            (default: :obj:`4.5`)
        max_z (int, optional): Maximum atomic number. Used for initializing embeddings.
            (default: :obj:`128`)
        max_num_neighbors (int, optional): Maximum number of neighbors to return for a
            given node/atom when constructing the molecular graph during forward passes.
            (default: :obj:`64`)
        equivariance_invariance_group (string, optional): Group under whose action on input
            positions internal tensor features will be equivariant and scalar predictions
            will be invariant. O(3) or SO(3).
            (default :obj:`"O(3)"`)
        box_vecs (Tensor, optional):
            The vectors defining the periodic box.  This must have shape `(3, 3)`,
            where `box_vectors[0] = a`, `box_vectors[1] = b`, and `box_vectors[2] = c`.
            If this is omitted, periodic boundary conditions are not applied.
            (default: :obj:`None`)
        static_shapes (bool, optional): Whether to enforce static shapes.
            Makes the model CUDA-graph compatible.
            (default: :obj:`True`)
    """

    def __init__(
        self,
        hidden_channels=128,
        num_layers=2,
        num_rbf=32,
        rbf_type="expnorm",
        trainable_rbf=False,
        activation="silu",
        cutoff_lower=0,
        cutoff_upper=4.5,
        max_num_neighbors=64,
        max_z=128,
        equivariance_invariance_group="O(3)",
        static_shapes=True,
        dtype=torch.float32,
        box_vecs=None,
    ):
        super(TensorNet, self).__init__()

        assert rbf_type in rbf_class_mapping, (
            f'Unknown RBF type "{rbf_type}". '
            f'Choose from {", ".join(rbf_class_mapping.keys())}.'
        )
        assert activation in act_class_mapping, (
            f'Unknown activation function "{activation}". '
            f'Choose from {", ".join(act_class_mapping.keys())}.'
        )

        assert equivariance_invariance_group in ["O(3)", "SO(3)"], (
            f'Unknown group "{equivariance_invariance_group}". '
            f"Choose O(3) or SO(3)."
        )
        self.hidden_channels = hidden_channels
        self.equivariance_invariance_group = equivariance_invariance_group
        self.num_layers = num_layers
        self.num_rbf = num_rbf
        self.rbf_type = rbf_type
        self.activation = activation
        self.cutoff_lower = cutoff_lower
        self.cutoff_upper = cutoff_upper
        act_class = act_class_mapping[activation]
        self.distance_expansion = rbf_class_mapping[rbf_type](
            cutoff_lower, cutoff_upper, num_rbf, trainable_rbf
        )
        self.tensor_embedding = TensorEmbedding(
            hidden_channels,
            num_rbf,
            act_class,
            cutoff_lower,
            cutoff_upper,
            trainable_rbf,
            max_z,
            dtype,
        )

        self.layers = nn.ModuleList()
        if num_layers != 0:
            for _ in range(num_layers):
                self.layers.append(
                    Interaction(
                        num_rbf,
                        hidden_channels,
                        act_class,
                        cutoff_lower,
                        cutoff_upper,
                        equivariance_invariance_group,
                        dtype,
                    )
                )
        self.linear = nn.Linear(3 * hidden_channels, hidden_channels, dtype=dtype)
        self.out_norm = nn.LayerNorm(3 * hidden_channels, dtype=dtype)
        self.act = act_class()
        # Resize to fit set to false ensures Distance returns a statically-shaped tensor of size max_num_pairs=pos.size*max_num_neigbors
        # negative max_num_pairs argument means "per particle"
        # long_edge_index set to False saves memory and spares some kernel launches by keeping neighbor indices as int32.
        self.static_shapes = static_shapes
        self.distance = OptimizedDistance(
            cutoff_lower,
            cutoff_upper,
            max_num_pairs=-max_num_neighbors,
            return_vecs=True,
            loop=True,
            resize_to_fit=not self.static_shapes,
            box=box_vecs,
            long_edge_index=True,
        )

        # for torchscript
        self.opt = OPT

        self.reset_parameters()

    def reset_parameters(self):
        self.tensor_embedding.reset_parameters()
        for layer in self.layers:
            layer.reset_parameters()
        self.linear.reset_parameters()
        self.out_norm.reset_parameters()

    def setup_for_inference(self, z, batch):
        self.tensor_embedding.setup_for_inference(z, batch)
        self.inference_mode = True

    def forward(
        self,
        z: Tensor,
        pos: Tensor,
        batch: Tensor,
        box: Optional[Tensor] = None,
        q: Optional[Tensor] = None,
        s: Optional[Tensor] = None,
    ) -> Tuple[Tensor, Optional[Tensor], Tensor, Tensor, Tensor]:
        # Obtain graph, with distances and relative position vectors
        edge_index, edge_weight, edge_vec = self.distance(pos, batch, box)

        if self.opt:
            # perpare graph indices for message passing
            row_data, row_indices, row_indptr, col_data, col_indices, col_indptr = (
                graph_transform(edge_index.int(), z.shape[0])
            )
        else:
            row_data, row_indices, row_indptr, col_data, col_indices, col_indptr = (
                torch.empty(0, dtype=torch.long),
                torch.empty(0, dtype=torch.long),
                torch.empty(0, dtype=torch.long),
                torch.empty(0, dtype=torch.long),
                torch.empty(0, dtype=torch.long),
                torch.empty(0, dtype=torch.long),
            )

        # This assert convinces TorchScript that edge_vec is a Tensor and not an Optional[Tensor]
        assert (
            edge_vec is not None
        ), "Distance module did not return directional information"
        # Distance module returns -1 for non-existing edges, to avoid having to resize the tensors when we want to ensure static shapes (for CUDA graphs) we make all non-existing edges pertain to a ghost atom
        # Total charge q is a molecule-wise property. We transform it into an atom-wise property, with all atoms belonging to the same molecule being assigned the same charge q
        if q is None:
            q = torch.zeros_like(z, device=z.device, dtype=z.dtype)
        else:
            q = q[batch]
        zp = z
        if self.static_shapes:
            mask = (edge_index[0] < 0).unsqueeze(0).expand_as(edge_index)
            zp = torch.cat((z, torch.zeros(1, device=z.device, dtype=z.dtype)), dim=0)

            if not self.opt:
                q = torch.cat(
                    (q, torch.zeros(1, device=q.device, dtype=q.dtype)), dim=0
                )
            # I trick the model into thinking that the masked edges pertain to the extra atom
            # WARNING: This can hurt performance if max_num_pairs >> actual_num_pairs
            edge_index = edge_index.masked_fill(mask, z.shape[0])
            edge_weight = edge_weight.masked_fill(mask[0], 0)
            edge_vec = edge_vec.masked_fill(
                mask[0].unsqueeze(-1).expand_as(edge_vec), 0
            )

        edge_attr = self.distance_expansion(edge_weight)
        mask = edge_index[0] == edge_index[1]
        # Normalizing edge vectors by their length can result in NaNs, breaking Autograd.
        # I avoid dividing by zero by setting the weight of self edges and self loops to 1
        edge_vec = edge_vec / edge_weight.masked_fill(mask, 1).unsqueeze(1)

        graph_info = {
            "edge_index": edge_index,
            "row_data": row_data,
            "row_indices": row_indices,
            "row_indptr": row_indptr,
            "col_data": col_data,
            "col_indices": col_indices,
            "col_indptr": col_indptr,
        }

        X = self.tensor_embedding(
            zp, graph_info, edge_weight, edge_vec, edge_attr
        )  # [N, 3, 3, F]
        for layer in self.layers:
            X = layer(X, graph_info, edge_weight, edge_attr, q)  # [N, 3, 3, F]

        if not self.opt:
            I, A, S = decompose_tensor(X)
            x = torch.cat((3 * I**2, tensor_norm(A), tensor_norm(S)), dim=-1)
            x = self.out_norm(x)
            x = self.act(self.linear((x)))

            # Remove the extra atom
            if self.static_shapes:
                x = x[:-1]

        else:

            x = fn_tensor_norm3(X)
            x = self.out_norm(x)
            x = self.act(self.linear((x)))

            # with the OPT code and static shapes there is no dummy atom in the outputs

        return x, None, z, pos, batch


def tensornet_embedding_message_passing(
    edge_vec_norm, edge_attr_processed, edge_index, num_atoms: int
):
    # written to mirror the optmized fn_tensornet_embedding_message_passing

    E, _, F = edge_attr_processed.shape

    Iij = edge_attr_processed[:, 0, :]  # [num_edges, hidden_channels]
    Aij = edge_attr_processed[:, 1, None, :] * edge_vec_norm.unsqueeze(
        -1
    )  # [num_edges, 3, hidden_channels]
    _outer = torch.matmul(
        edge_vec_norm.unsqueeze(2), edge_vec_norm.unsqueeze(-2)
    )  # [num_edges, 3, 3]
    Sij = edge_attr_processed[:, 2, None, None, :] * _outer.unsqueeze(
        -1
    )  # [num_edges, 3, 3, hidden_channels]

    source_scalar = torch.zeros(
        num_atoms, F, device=edge_attr_processed.device, dtype=Iij.dtype
    )
    source_vector = torch.zeros(
        num_atoms, 3, F, device=edge_attr_processed.device, dtype=Iij.dtype
    )
    source_tensor = torch.zeros(
        num_atoms, 3, 3, F, device=edge_attr_processed.device, dtype=Iij.dtype
    )
    I = source_scalar.index_add(
        dim=0, index=edge_index[0], source=Iij
    )  # [num_atoms, hidden_channels]
    A_vec = source_vector.index_add(
        dim=0, index=edge_index[0], source=Aij
    )  # [num_atoms, 3, hidden_channels]
    S = source_tensor.index_add(
        dim=0, index=edge_index[0], source=Sij
    )  # [num_atoms, 3, 3, hidden_channels]

    A = vector_to_skewtensor(A_vec)  # [num_atoms, 3, 3, hidden_channels]
    S = outer_to_symtensor(S)  # [num_atoms, 3, 3, hidden_channels]

    return I, A, S


class TensorEmbedding(nn.Module):
    """Tensor embedding layer.

    :meta private:
    """

    def __init__(
        self,
        hidden_channels,
        num_rbf,
        activation,
        cutoff_lower,
        cutoff_upper,
        trainable_rbf=False,
        max_z=128,
        dtype=torch.float32,
    ):
        super(TensorEmbedding, self).__init__()

        self.hidden_channels = hidden_channels
        self.distance_proj1 = nn.Linear(num_rbf, hidden_channels, dtype=dtype)
        self.distance_proj2 = nn.Linear(num_rbf, hidden_channels, dtype=dtype)
        self.distance_proj3 = nn.Linear(num_rbf, hidden_channels, dtype=dtype)
        self.cutoff = CosineCutoff(cutoff_lower, cutoff_upper)
        self.max_z = max_z
        self.emb = nn.Embedding(max_z, hidden_channels, dtype=dtype)
        self.emb2 = nn.Linear(2 * hidden_channels, hidden_channels, dtype=dtype)
        self.act = activation()
        self.linears_tensor = nn.ModuleList()
        for _ in range(3):
            self.linears_tensor.append(
                nn.Linear(hidden_channels, hidden_channels, bias=False)
            )
        self.linears_scalar = nn.ModuleList()
        self.linears_scalar.append(
            nn.Linear(hidden_channels, 2 * hidden_channels, bias=True, dtype=dtype)
        )
        self.linears_scalar.append(
            nn.Linear(2 * hidden_channels, 3 * hidden_channels, bias=True, dtype=dtype)
        )
        self.init_norm = nn.LayerNorm(hidden_channels, dtype=dtype)

        # for torchscript compatibility
        self.Zij_map = torch.empty(0, 0, 0)
        self.opt = OPT

        self.reset_parameters()

    def reset_parameters(self):
        self.distance_proj1.reset_parameters()
        self.distance_proj2.reset_parameters()
        self.distance_proj3.reset_parameters()
        self.emb.reset_parameters()
        self.emb2.reset_parameters()
        for linear in self.linears_tensor:
            linear.reset_parameters()
        for linear in self.linears_scalar:
            linear.reset_parameters()
        self.init_norm.reset_parameters()
        self.inference_mode = False

    def setup_for_inference(self, z, batch):

        # we can precompute the edge-wise atomic embeddings
        with torch.no_grad():
            max_z = z.max().item()
            _zij_map = torch.zeros((max_z + 1, max_z + 1, self.hidden_channels))

            for zi in torch.unique(z):
                Zi = self.emb(zi)
                for zj in torch.unique(z):
                    Zj = self.emb(zj)
                    _zij_map[zi, zj] = self.emb2(torch.cat([Zi, Zj], dim=-1))

        del self.Zij_map
        self.register_buffer("Zij_map", _zij_map)
        self.inference_mode = True

    def _get_atomic_number_message(self, z: Tensor, edge_index: Tensor) -> Tensor:

        if self.inference_mode:
            if torch.jit.is_scripting():
                # torchscript does not like the buffer
                Zij = torch.as_tensor(self.Zij_map)[z[edge_index[0]], z[edge_index[1]]]
            else:
                Zij = self.Zij_map[z[edge_index[0]], z[edge_index[1]]]
        else:
            Z = self.emb(z)
            Zij = self.emb2(
                Z.index_select(0, edge_index.t().reshape(-1)).view(
                    -1, self.hidden_channels * 2
                )
            )
        return Zij

    def forward(
        self,
        z: Tensor,
        graph_info: dict[str, Tensor],
        edge_weight: Tensor,
        edge_vec_norm: Tensor,
        edge_attr: Tensor,
    ) -> Tensor:

        edge_index = graph_info["edge_index"]

        Zij = self._get_atomic_number_message(
            z, edge_index
        )  # [num_edges, hidden_channels]

        dp1 = self.distance_proj1(edge_attr)  # [num_edges, hidden_channels]
        dp2 = self.distance_proj2(edge_attr)  # [num_edges, hidden_channels]
        dp3 = self.distance_proj3(edge_attr)  # [num_edges, hidden_channels]
        CZij = (
            self.cutoff(edge_weight).unsqueeze(-1) * Zij
        )  # [num_edges, hidden_channels]

        edge_attr_processed = CZij.unsqueeze(1) * torch.stack(
            [dp1, dp2, dp3], dim=1
        )  # [num_edges, 3, hidden_channels]

        if not self.opt:
            I, A, S = tensornet_embedding_message_passing(
                edge_vec_norm, edge_attr_processed, edge_index, z.shape[0]
            )  # [num_atoms, hidden_channels], [num_atoms, 3, 3, hidden_channels], [num_atoms, 3, 3, hidden_channels]

        else:
            row_indptr, row_indices, row_data = (
                graph_info["row_indptr"],
                graph_info["row_indices"],
                graph_info["row_data"],
            )
            I, A, S = fn_tensornet_embedding_message_passing(
                edge_vec_norm,
                edge_attr_processed,
                row_data,
                row_indptr,
            )  # [num_atoms, 1, hidden_channels], [num_atoms, 3, hidden_channels], [num_atoms, 5, hidden_channels]

        X = compose_tensor(I, A, S)  # [num_atoms, 3, 3, hidden_channels]

        norm = self.init_norm(tensor_norm(X))
        for linear_scalar in self.linears_scalar:
            norm = self.act(linear_scalar(norm))

        norm = norm.reshape(-1, 3, self.hidden_channels)

        if not self.opt:
            I = (
                self.linears_tensor[0](I) * norm[:, 0, :]
            )  # [num_atoms, hidden_channels]
            A = (
                self.linears_tensor[1](A) * norm[:, 1, None, None, :]
            )  # [num_atoms, 3, 3, hidden_channels]
            S = (
                self.linears_tensor[2](S) * norm[:, 2, None, None, :]
            )  # [num_atoms, 3, 3, hidden_channels]

        else:
            I = (
                self.linears_tensor[0](I) * norm[:, 0, None, :]
            )  # [num_atoms, 1, hidden_channels]
            A = (
                self.linears_tensor[1](A) * norm[:, 1, None, :]
            )  # [num_atoms, 3, hidden_channels]
            S = (
                self.linears_tensor[2](S) * norm[:, 2, None, :]
            )  # [num_atoms, 5, hidden_channels]

        X = compose_tensor(I, A, S)  # [num_atoms, 3, 3, hidden_channels]

        return X


def tensornet_interaction_message_passing(
    I, A, S, edge_attr_processed, edge_index, natoms: int
):

    # written to mirror the optimized fn_tensornet_interaction_message_passing

    A_vec = skewtensor_to_vector(A)  # [num_atoms, 3, hidden_channels]

    factor_scalar = edge_attr_processed[..., 0, :]
    factor_vector = edge_attr_processed[..., 1, None, :]
    factor_tensor = edge_attr_processed[..., 2, None, None, :]

    I = scalar_message_passing(
        edge_index, factor_scalar, I, natoms
    )  # [num_atoms, hidden_channels]
    A_vec = vector_message_passing(
        edge_index, factor_vector, A_vec, natoms
    )  # [num_atoms, 3, hidden_channels]
    S = tensor_message_passing(
        edge_index, factor_tensor, S, natoms
    )  # [num_atoms, 3, 3, hidden_channels]

    A = vector_to_skewtensor(A_vec)  # [num_atoms, 3, 3, hidden_channels]
    return I, A, S


def scalar_message_passing(
    edge_index: Tensor, factor: Tensor, tensor: Tensor, natoms: int
) -> Tensor:
    msg = factor * tensor.index_select(0, edge_index[1])  # [num_edges, hidden_channels]
    shape = (natoms, tensor.shape[1])
    tensor_m = torch.zeros(*shape, device=tensor.device, dtype=tensor.dtype)
    tensor_m = tensor_m.index_add(0, edge_index[0], msg)
    return tensor_m


def vector_message_passing(
    edge_index: Tensor, factor: Tensor, tensor: Tensor, natoms: int
) -> Tensor:
    msg = factor * tensor.index_select(
        0, edge_index[1]
    )  # [num_edges, 3, hidden_channels]
    shape = (natoms, tensor.shape[1], tensor.shape[2])
    tensor_m = torch.zeros(*shape, device=tensor.device, dtype=tensor.dtype)
    tensor_m = tensor_m.index_add(0, edge_index[0], msg)
    return tensor_m


def tensor_message_passing(
    edge_index: Tensor, factor: Tensor, tensor: Tensor, natoms: int
) -> Tensor:
    msg = factor * tensor.index_select(
        0, edge_index[1]
    )  # [num_edges, 3, 3, hidden_channels]
    shape = (natoms, tensor.shape[1], tensor.shape[2], tensor.shape[3])
    tensor_m = torch.zeros(*shape, device=tensor.device, dtype=tensor.dtype)
    tensor_m = tensor_m.index_add(0, edge_index[0], msg)
    return tensor_m


class Interaction(nn.Module):
    """Interaction layer.

    :meta private:
    """

    def __init__(
        self,
        num_rbf,
        hidden_channels,
        activation,
        cutoff_lower,
        cutoff_upper,
        equivariance_invariance_group,
        dtype=torch.float32,
    ):
        super(Interaction, self).__init__()

        self.num_rbf = num_rbf
        self.hidden_channels = hidden_channels
        self.cutoff = CosineCutoff(cutoff_lower, cutoff_upper)
        self.linears_scalar = nn.ModuleList()
        self.linears_scalar.append(
            nn.Linear(num_rbf, hidden_channels, bias=True, dtype=dtype)
        )
        self.linears_scalar.append(
            nn.Linear(hidden_channels, 2 * hidden_channels, bias=True, dtype=dtype)
        )
        self.linears_scalar.append(
            nn.Linear(2 * hidden_channels, 3 * hidden_channels, bias=True, dtype=dtype)
        )
        self.linears_tensor = nn.ModuleList()
        for _ in range(6):
            self.linears_tensor.append(
                nn.Linear(hidden_channels, hidden_channels, bias=False)
            )
        self.act = activation()
        self.equivariance_invariance_group = equivariance_invariance_group
        self.opt = OPT
        self.reset_parameters()

    def reset_parameters(self):
        for linear in self.linears_scalar:
            linear.reset_parameters()
        for linear in self.linears_tensor:
            linear.reset_parameters()

    def forward(
        self,
        X: Tensor,
        graph_info: dict[str, Tensor],
        edge_weight: Tensor,
        edge_attr: Tensor,
        q: Tensor,
    ) -> Tensor:

        C = self.cutoff(edge_weight)
        for linear_scalar in self.linears_scalar:
            edge_attr = self.act(linear_scalar(edge_attr))
        edge_attr = (edge_attr * C.view(-1, 1)).reshape(
            edge_attr.shape[0], 3, self.hidden_channels
        )

        X = X / (tensor_norm(X) + 1)[:, None, None, :]

        I, A, S = decompose_tensor(
            X
        )  # ([num_atoms, hidden_channels], [num_atoms, 3, 3, hidden_channels], [num_atoms, 3, 3, hidden_channels])
        # or ([num_atoms, 1, hidden_channels], [num_atoms, 3, hidden_channels], [num_atoms, 5, hidden_channels])

        I = self.linears_tensor[0](I)
        A = self.linears_tensor[1](A)
        S = self.linears_tensor[2](S)
        Y = compose_tensor(I, A, S)  # [num_atoms, 3, 3, hidden_channels]

        if not self.opt:
            edge_index = graph_info["edge_index"]

            Im, Am, Sm = tensornet_interaction_message_passing(
                I, A, S, edge_attr, edge_index, X.shape[0]
            )  # ([num_atoms, hidden_channels], [num_atoms, 3, 3, hidden_channels], [num_atoms, 3, 3, hidden_channels])

        else:

            row_data = graph_info["row_data"]
            row_indices = graph_info["row_indices"]
            row_indptr = graph_info["row_indptr"]
            col_data = graph_info["col_data"]
            col_indices = graph_info["col_indices"]
            col_indptr = graph_info["col_indptr"]

            Im, Am, Sm = fn_tensornet_interaction_message_passing(
                I,
                A,
                S,
                edge_attr,
                col_data,
                col_indices,
                col_indptr,
                row_data,
                row_indices,
                row_indptr,
            )  # ([num_atoms, 1, hidden_channels], [num_atoms, 3, hidden_channels], [num_atoms, 5, hidden_channels])

        msg = compose_tensor(Im, Am, Sm)  # [num_atoms, 3, 3, hidden_channels]

        if self.equivariance_invariance_group == "O(3)":
            C = (1 + 0.1 * q[..., None, None, None]) * tensor_matmul_o3(Y, msg)
            I, A, S = decompose_tensor(C)
        if self.equivariance_invariance_group == "SO(3)":
            C = 2 * tensor_matmul_so3(Y, msg)
            I, A, S = decompose_tensor(C)

        normp1 = tensor_norm(C) + 1

        if not self.opt:
            I, A, S = (
                I / normp1,
                A / normp1[..., None, None, :],
                S / normp1[..., None, None, :],
            )
        else:
            I = I / normp1.unsqueeze(1)
            A = A / normp1.unsqueeze(1)
            S = S / normp1.unsqueeze(1)

        I = self.linears_tensor[3](I)
        A = self.linears_tensor[4](A)
        S = self.linears_tensor[5](S)
        dX = compose_tensor(I, A, S)
        X = X + dX + (1 + 0.1 * q[..., None, None, None]) * tensor_matmul_so3(dX, dX)

        return X



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import warp as wp

from .utils import add_module, get_wp_fp_dtype


def generate_radial_message_passing(dtype: str):
    dtype_wp = get_wp_fp_dtype(dtype)

    class vec3(wp.types.vector(length=3, dtype=dtype_wp)):
        pass

    class vec5(wp.types.vector(length=5, dtype=dtype_wp)):
        pass

    def radial_message_passing_fwd(
        edge_vec_norm: wp.array(ndim=2, dtype=dtype_wp),
        edge_attr: wp.array(ndim=3, dtype=dtype_wp),
        row_data: wp.array(ndim=1, dtype=wp.int32),
        row_indptr: wp.array(ndim=1, dtype=wp.int32),
        output_I: wp.array(ndim=3, dtype=dtype_wp),
        output_A: wp.array(ndim=3, dtype=dtype_wp),
        output_S: wp.array(ndim=3, dtype=dtype_wp),
    ):
        b, h = wp.tid()

        output_I_reg = output_I.dtype(0)
        output_A_reg = vec3(output_I.dtype(0))
        output_S_reg = vec5(output_I.dtype(0))

        for i in range(row_indptr[b], row_indptr[b + 1]):
            idx_w = row_data[i]

            weight_I_reg = edge_attr[idx_w, 0, h]
            weight_A_reg = edge_attr[idx_w, 1, h]
            weight_S_reg = edge_attr[idx_w, 2, h]

            r_ij = vec3(output_I.dtype(0))
            r_ij[0] = edge_vec_norm[idx_w, 0]
            r_ij[1] = edge_vec_norm[idx_w, 1]
            r_ij[2] = edge_vec_norm[idx_w, 2]

            output_I_reg += weight_I_reg

            output_A_reg[0] += r_ij[2] * weight_A_reg
            output_A_reg[1] += -r_ij[1] * weight_A_reg
            output_A_reg[2] += r_ij[0] * weight_A_reg

            S_reg = vec5()
            mean_r2 = (r_ij[0] * r_ij[0] + r_ij[1] * r_ij[1] + r_ij[2] * r_ij[2]) / output_I.dtype(3.0)
            S_reg[0] = r_ij[0] * r_ij[0] - mean_r2
            S_reg[1] = r_ij[0] * r_ij[1]
            S_reg[2] = r_ij[0] * r_ij[2]
            S_reg[3] = r_ij[1] * r_ij[1] - mean_r2
            S_reg[4] = r_ij[1] * r_ij[2]

            output_S_reg[0] += S_reg[0] * weight_S_reg
            output_S_reg[1] += S_reg[1] * weight_S_reg
            output_S_reg[2] += S_reg[2] * weight_S_reg
            output_S_reg[3] += S_reg[3] * weight_S_reg
            output_S_reg[4] += S_reg[4] * weight_S_reg

        output_I[b, 0, h] = output_I_reg
        for i in range(3):
            output_A[b, i, h] = output_A_reg[i]

        for i in range(5):
            output_S[b, i, h] = output_S_reg[i]

    def radial_message_passing_bwd(
        edge_vec_norm: wp.array(ndim=2, dtype=dtype_wp),
        edge_attr: wp.array(ndim=3, dtype=dtype_wp),
        row_data: wp.array(ndim=1, dtype=wp.int32),
        row_indptr: wp.array(ndim=1, dtype=wp.int32),
        doutput_I: wp.array(ndim=3, dtype=dtype_wp),
        doutput_A: wp.array(ndim=3, dtype=dtype_wp),
        doutput_S: wp.array(ndim=3, dtype=dtype_wp),
        dedge_vec_norm: wp.array(ndim=2, dtype=dtype_wp),
        dedge_attr: wp.array(ndim=3, dtype=dtype_wp),
    ):
        b, h = wp.tid()

        doutput_I_reg = doutput_I[b, 0, h]
        doutput_A_reg = vec3()
        doutput_A_reg[0] = doutput_A[b, 0, h]
        doutput_A_reg[1] = doutput_A[b, 1, h]
        doutput_A_reg[2] = doutput_A[b, 2, h]

        doutput_S_reg = vec5()
        doutput_S_reg[0] = doutput_S[b, 0, h]
        doutput_S_reg[1] = doutput_S[b, 1, h]
        doutput_S_reg[2] = doutput_S[b, 2, h]
        doutput_S_reg[3] = doutput_S[b, 3, h]
        doutput_S_reg[4] = doutput_S[b, 4, h]

        for i in range(row_indptr[b], row_indptr[b + 1]):
            idx_w = row_data[i]

            edge_attr_A_reg = edge_attr[idx_w, 1, h]
            edge_attr_S_reg = edge_attr[idx_w, 2, h]

            r_ij = vec3(doutput_I.dtype(0))
            dr_ij = vec3(doutput_I.dtype(0))
            r_ij[0] = edge_vec_norm[idx_w, 0]
            r_ij[1] = edge_vec_norm[idx_w, 1]
            r_ij[2] = edge_vec_norm[idx_w, 2]

            dr_ij[2] += doutput_A_reg[0] * edge_attr_A_reg
            dr_ij[1] += -doutput_A_reg[1] * edge_attr_A_reg
            dr_ij[0] += doutput_A_reg[2] * edge_attr_A_reg

            dedge_attr_I = doutput_I_reg

            dedge_attr_A = doutput_A_reg[0] * r_ij[2] - doutput_A_reg[1] * r_ij[1] + doutput_A_reg[2] * r_ij[0]

            S_reg = vec5()
            mean_r2 = (r_ij[0] * r_ij[0] + r_ij[1] * r_ij[1] + r_ij[2] * r_ij[2]) / doutput_I.dtype(3.0)
            S_reg[0] = r_ij[0] * r_ij[0] - mean_r2
            S_reg[1] = r_ij[0] * r_ij[1]
            S_reg[2] = r_ij[0] * r_ij[2]
            S_reg[3] = r_ij[1] * r_ij[1] - mean_r2
            S_reg[4] = r_ij[1] * r_ij[2]

            dedge_attr_S = (S_reg[0]) * doutput_S_reg[0]
            dedge_attr_S += (S_reg[1]) * doutput_S_reg[1]
            dedge_attr_S += (S_reg[2]) * doutput_S_reg[2]
            dedge_attr_S += (S_reg[3]) * doutput_S_reg[3]
            dedge_attr_S += (S_reg[4]) * doutput_S_reg[4]

            dS_reg = vec5()
            dS_reg[0] = edge_attr_S_reg * doutput_S_reg[0]
            dS_reg[1] = edge_attr_S_reg * doutput_S_reg[1]
            dS_reg[2] = edge_attr_S_reg * doutput_S_reg[2]
            dS_reg[3] = edge_attr_S_reg * doutput_S_reg[3]
            dS_reg[4] = edge_attr_S_reg * doutput_S_reg[4]

            dr_ij[0] += (
                dS_reg[0] * (doutput_I.dtype(4.0) / doutput_I.dtype(3.0) * r_ij[0])
                + dS_reg[1] * r_ij[1]
                + dS_reg[2] * r_ij[2]
                + dS_reg[3] * (-doutput_I.dtype(2.0) / doutput_I.dtype(3.0) * r_ij[0])
            )
            dr_ij[1] += (
                dS_reg[0] * (-doutput_I.dtype(2.0) / doutput_I.dtype(3.0) * r_ij[1])
                + dS_reg[1] * r_ij[0]
                + dS_reg[3] * (doutput_I.dtype(4.0) / doutput_I.dtype(3.0) * r_ij[1])
                + dS_reg[4] * r_ij[2]
            )
            dr_ij[2] += (
                dS_reg[0] * (-doutput_I.dtype(2.0) / doutput_I.dtype(3.0) * r_ij[2])
                + dS_reg[2] * r_ij[0]
                + dS_reg[3] * (-doutput_I.dtype(2.0) / doutput_I.dtype(3.0) * r_ij[2])
                + dS_reg[4] * r_ij[1]
            )

            wp.atomic_add(dedge_attr, idx_w, 0, h, dedge_attr_I)
            wp.atomic_add(dedge_attr, idx_w, 1, h, dedge_attr_A)
            wp.atomic_add(dedge_attr, idx_w, 2, h, dedge_attr_S)

            wp.atomic_add(dedge_vec_norm, idx_w, 0, dr_ij[0])
            wp.atomic_add(dedge_vec_norm, idx_w, 1, dr_ij[1])
            wp.atomic_add(dedge_vec_norm, idx_w, 2, dr_ij[2])

    def radial_message_passing_bwd_bwd(
        edge_vec_norm: wp.array(ndim=2, dtype=dtype_wp),
        edge_attr: wp.array(ndim=3, dtype=dtype_wp),
        dedge_vec_norm: wp.array(ndim=2, dtype=dtype_wp),
        dedge_attr: wp.array(ndim=3, dtype=dtype_wp),
        doutput_I: wp.array(ndim=3, dtype=dtype_wp),
        doutput_A: wp.array(ndim=3, dtype=dtype_wp),
        doutput_S: wp.array(ndim=3, dtype=dtype_wp),
        row_data: wp.array(ndim=1, dtype=wp.int32),
        row_indptr: wp.array(ndim=1, dtype=wp.int32),
        d2edge_vec_norm: wp.array(ndim=2, dtype=dtype_wp),
        d2edge_attr: wp.array(ndim=3, dtype=dtype_wp),
        d2output_I: wp.array(ndim=3, dtype=dtype_wp),
        d2output_A: wp.array(ndim=3, dtype=dtype_wp),
        d2output_S: wp.array(ndim=3, dtype=dtype_wp),
    ):
        b, h = wp.tid()

        d2output_I_reg = d2output_I.dtype(0.0)
        d2output_A_reg = vec3()
        d2output_S_reg = vec5()

        for i in range(row_indptr[b], row_indptr[b + 1]):
            idx_w = row_data[i]
            edge_attr_A_reg = edge_attr[idx_w, 1, h]
            edge_attr_S_reg = edge_attr[idx_w, 2, h]

            dedge_attr_I = dedge_attr[idx_w, 0, h]
            dedge_attr_A = dedge_attr[idx_w, 1, h]
            dedge_attr_S = dedge_attr[idx_w, 2, h]

            r_ij = vec3(d2output_I.dtype(0))
            dr_ij = vec3(d2output_I.dtype(0))
            for j in range(3):
                r_ij[j] = edge_vec_norm[idx_w, j]
                dr_ij[j] = dedge_vec_norm[idx_w, j]

            d2output_I_reg += dedge_attr_I

            d2r_ij = vec3(d2output_I.dtype(0))

            # No gradient contribution for edge_attr[*, 0, h] in forward pass
            # d2edge_attr[idx_w, 0, h] = d2output_I.dtype(0.0)

            d2output_A_reg[0] += dr_ij[2] * edge_attr_A_reg
            d2output_A_reg[1] += -dr_ij[1] * edge_attr_A_reg
            d2output_A_reg[2] += dr_ij[0] * edge_attr_A_reg

            d2output_A_reg[0] += dedge_attr_A * r_ij[2]
            d2output_A_reg[1] += -dedge_attr_A * r_ij[1]
            d2output_A_reg[2] += dedge_attr_A * r_ij[0]

            dweight_A = doutput_A[b, 0, h] * dr_ij[2] - doutput_A[b, 1, h] * dr_ij[1] + doutput_A[b, 2, h] * dr_ij[0]

            d2r_ij[2] += dedge_attr_A * doutput_A[b, 0, h]
            d2r_ij[1] += -dedge_attr_A * doutput_A[b, 1, h]
            d2r_ij[0] += dedge_attr_A * doutput_A[b, 2, h]

            wp.atomic_add(d2edge_attr, idx_w, 1, h, dweight_A)

            c0 = doutput_S.dtype(4.0) / doutput_S.dtype(3.0)
            c1 = -doutput_S.dtype(2.0) / doutput_S.dtype(3.0)

            c2 = doutput_S.dtype(2.0) / doutput_S.dtype(3.0)
            c3 = -doutput_S.dtype(1.0) / doutput_S.dtype(3.0)

            d2output_S_reg[0] += edge_attr_S_reg * (
                dedge_vec_norm[idx_w, 0] * c0 * r_ij[0]
                + dedge_vec_norm[idx_w, 1] * c1 * r_ij[1]
                + dedge_vec_norm[idx_w, 2] * c1 * r_ij[2]
            )
            d2output_S_reg[0] += dedge_attr_S * (
                c2 * r_ij[0] * r_ij[0] + c3 * r_ij[1] * r_ij[1] + c3 * r_ij[2] * r_ij[2]
            )

            d2output_S_reg[1] += edge_attr_S_reg * (
                dedge_vec_norm[idx_w, 0] * r_ij[1] + dedge_vec_norm[idx_w, 1] * r_ij[0]
            )
            d2output_S_reg[1] += dedge_attr_S * (r_ij[1] * r_ij[0])

            d2output_S_reg[2] += edge_attr_S_reg * (
                dedge_vec_norm[idx_w, 0] * r_ij[2] + dedge_vec_norm[idx_w, 2] * r_ij[0]
            )
            d2output_S_reg[2] += dedge_attr_S * (r_ij[2] * r_ij[0])

            d2output_S_reg[3] += edge_attr_S_reg * (
                dedge_vec_norm[idx_w, 0] * c1 * r_ij[0]
                + dedge_vec_norm[idx_w, 1] * c0 * r_ij[1]
                + dedge_vec_norm[idx_w, 2] * c1 * r_ij[2]
            )
            d2output_S_reg[3] += dedge_attr_S * (
                c3 * r_ij[0] * r_ij[0] + c2 * r_ij[1] * r_ij[1] + c3 * r_ij[2] * r_ij[2]
            )

            d2output_S_reg[4] += edge_attr_S_reg * (
                dedge_vec_norm[idx_w, 1] * r_ij[2] + dedge_vec_norm[idx_w, 2] * r_ij[1]
            )
            d2output_S_reg[4] += dedge_attr_S * (r_ij[2] * r_ij[1])

            d2r_ij[0] += doutput_S[b, 0, h] * edge_attr_S_reg * (dedge_vec_norm[idx_w, 0] * c0)
            d2r_ij[1] += doutput_S[b, 0, h] * edge_attr_S_reg * (dedge_vec_norm[idx_w, 1] * c1)
            d2r_ij[2] += doutput_S[b, 0, h] * edge_attr_S_reg * (dedge_vec_norm[idx_w, 2] * c1)

            d2r_ij[0] += doutput_S[b, 0, h] * dedge_attr_S * (c0 * r_ij[0])
            d2r_ij[1] += doutput_S[b, 0, h] * dedge_attr_S * (c1 * r_ij[1])
            d2r_ij[2] += doutput_S[b, 0, h] * dedge_attr_S * (c1 * r_ij[2])

            d2r_ij[0] += doutput_S[b, 1, h] * edge_attr_S_reg * (dedge_vec_norm[idx_w, 1])
            d2r_ij[1] += doutput_S[b, 1, h] * edge_attr_S_reg * (dedge_vec_norm[idx_w, 0])

            d2r_ij[0] += doutput_S[b, 1, h] * dedge_attr_S * (r_ij[1])
            d2r_ij[1] += doutput_S[b, 1, h] * dedge_attr_S * (r_ij[0])

            d2r_ij[0] += doutput_S[b, 2, h] * edge_attr_S_reg * (dedge_vec_norm[idx_w, 2])
            d2r_ij[2] += doutput_S[b, 2, h] * edge_attr_S_reg * (dedge_vec_norm[idx_w, 0])

            d2r_ij[0] += doutput_S[b, 2, h] * dedge_attr_S * (r_ij[2])
            d2r_ij[2] += doutput_S[b, 2, h] * dedge_attr_S * (r_ij[0])

            d2r_ij[0] += doutput_S[b, 3, h] * edge_attr_S_reg * (dedge_vec_norm[idx_w, 0] * c1)
            d2r_ij[1] += doutput_S[b, 3, h] * edge_attr_S_reg * (dedge_vec_norm[idx_w, 1] * c0)
            d2r_ij[2] += doutput_S[b, 3, h] * edge_attr_S_reg * (dedge_vec_norm[idx_w, 2] * c1)

            d2r_ij[0] += doutput_S[b, 3, h] * dedge_attr_S * (c1 * r_ij[0])
            d2r_ij[1] += doutput_S[b, 3, h] * dedge_attr_S * (c0 * r_ij[1])
            d2r_ij[2] += doutput_S[b, 3, h] * dedge_attr_S * (c1 * r_ij[2])

            d2r_ij[1] += doutput_S[b, 4, h] * edge_attr_S_reg * (dedge_vec_norm[idx_w, 2])
            d2r_ij[2] += doutput_S[b, 4, h] * edge_attr_S_reg * (dedge_vec_norm[idx_w, 1])

            d2r_ij[1] += doutput_S[b, 4, h] * dedge_attr_S * (r_ij[2])
            d2r_ij[2] += doutput_S[b, 4, h] * dedge_attr_S * (r_ij[1])

            d2weight_S = doutput_S.dtype(0.0)
            d2weight_S += doutput_S[b, 0, h] * (
                c0 * r_ij[0] * dedge_vec_norm[idx_w, 0]
                + c1 * r_ij[1] * dedge_vec_norm[idx_w, 1]
                + c1 * r_ij[2] * dedge_vec_norm[idx_w, 2]
            )

            d2weight_S += doutput_S[b, 1, h] * (r_ij[1] * dedge_vec_norm[idx_w, 0] + r_ij[0] * dedge_vec_norm[idx_w, 1])

            d2weight_S += doutput_S[b, 2, h] * (r_ij[2] * dedge_vec_norm[idx_w, 0] + r_ij[0] * dedge_vec_norm[idx_w, 2])

            d2weight_S += doutput_S[b, 3, h] * (
                c1 * r_ij[0] * dedge_vec_norm[idx_w, 0]
                + c0 * r_ij[1] * dedge_vec_norm[idx_w, 1]
                + c1 * r_ij[2] * dedge_vec_norm[idx_w, 2]
            )

            d2weight_S += doutput_S[b, 4, h] * (r_ij[2] * dedge_vec_norm[idx_w, 1] + r_ij[1] * dedge_vec_norm[idx_w, 2])

            wp.atomic_add(d2edge_attr, idx_w, 2, h, d2weight_S)

            wp.atomic_add(d2edge_vec_norm, idx_w, 0, d2r_ij[0])
            wp.atomic_add(d2edge_vec_norm, idx_w, 1, d2r_ij[1])
            wp.atomic_add(d2edge_vec_norm, idx_w, 2, d2r_ij[2])

        d2output_I[b, 0, h] = d2output_I_reg

        for i in range(3):
            d2output_A[b, i, h] = d2output_A_reg[i]

        for i in range(5):
            d2output_S[b, i, h] = d2output_S_reg[i]

    return (
        wp.Kernel(
            radial_message_passing_fwd,
            key=f"radial_message_passing_fwd_{dtype}",
            module=wp.get_module(f"radial_message_passing_fwd_{dtype}"),
        ),
        wp.Kernel(
            radial_message_passing_bwd,
            key=f"radial_message_passing_bwd_{dtype}",
            module=wp.get_module(f"radial_message_passing_bwd_{dtype}"),
        ),
        wp.Kernel(
            radial_message_passing_bwd_bwd,
            key=f"radial_message_passing_bwd_bwd_{dtype}",
            module=wp.get_module(f"radial_message_passing_bwd_bwd_{dtype}"),
        ),
    )


(
    radial_message_passing_fwd_fp64,
    radial_message_passing_bwd_fp64,
    radial_message_passing_bwd_bwd_fp64,
) = generate_radial_message_passing("float64")
(
    radial_message_passing_fwd_fp32,
    radial_message_passing_bwd_fp32,
    radial_message_passing_bwd_bwd_fp32,
) = generate_radial_message_passing("float32")
(
    radial_message_passing_fwd_fp16,
    radial_message_passing_bwd_fp16,
    radial_message_passing_bwd_bwd_fp16,
) = generate_radial_message_passing("float16")

add_module("radial_message_passing_fwd", ["float64"], radial_message_passing_fwd_fp64)
add_module("radial_message_passing_bwd", ["float64"], radial_message_passing_bwd_fp64)
add_module("radial_message_passing_bwd_bwd", ["float64"], radial_message_passing_bwd_bwd_fp64)

add_module("radial_message_passing_fwd", ["float32"], radial_message_passing_fwd_fp32)
add_module("radial_message_passing_bwd", ["float32"], radial_message_passing_bwd_fp32)
add_module("radial_message_passing_bwd_bwd", ["float32"], radial_message_passing_bwd_bwd_fp32)

add_module("radial_message_passing_fwd", ["float16"], radial_message_passing_fwd_fp16)
add_module("radial_message_passing_bwd", ["float16"], radial_message_passing_bwd_fp16)
add_module("radial_message_passing_bwd_bwd", ["float16"], radial_message_passing_bwd_bwd_fp16)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

import torch
import warp as wp
from torch import Tensor

from torchmdnet.extensions.warp_kernels import get_module, get_stream


@torch.library.custom_op(
    "tensornet::tensor_matmul_o3_3x3_fwd_primitive",
    mutates_args=(),
    device_types=["cpu", "cuda"],
)
def _(x: Tensor, y: Tensor) -> Tensor:
    if x.shape[1] != 3 or x.shape[2] != 3 or y.shape[1] != 3 or y.shape[2] != 3:
        raise ValueError("x and y must be 3x3 matrices")
    if x.ndim != 4 or y.ndim != 4:
        raise ValueError("x and y must be 4D tensors")

    stream = get_stream(x.device)
    device = wp.device_from_torch(x.device)
    output = torch.empty_like(x)

    x_wp = wp.from_torch(x.detach(), return_ctype=True)
    y_wp = wp.from_torch(y.detach(), return_ctype=True)
    output_wp = wp.from_torch(output.detach(), return_ctype=True)

    tensor_matmul_o3_3x3_fwd = get_module("tensor_matmul_o3_3x3_fwd", [str(x.dtype)])
    wp.launch(
        tensor_matmul_o3_3x3_fwd,
        dim=(x.shape[0], x.shape[-1]),
        stream=stream,
        device=device,
        inputs=(x_wp, y_wp, output_wp),
    )

    return output


@torch.library.register_fake("tensornet::tensor_matmul_o3_3x3_fwd_primitive")
def _(x: Tensor, y: Tensor) -> Tensor:
    return torch.empty_like(x)


@torch.library.custom_op(
    "tensornet::tensor_matmul_o3_3x3_bwd_primitive",
    mutates_args=(),
    device_types=["cpu", "cuda"],
)
def _(grad_output: Tensor, x: Tensor, y: Tensor) -> list[Tensor]:
    if x.shape[1] != 3 or x.shape[2] != 3 or y.shape[1] != 3 or y.shape[2] != 3:
        raise ValueError("x and y must be 3x3 matrices")
    if x.ndim != 4 or y.ndim != 4:
        raise ValueError("x and y must be 4D tensors")

    stream = get_stream(x.device)
    device = wp.device_from_torch(x.device)
    grad_x = torch.empty_like(x)
    grad_y = torch.empty_like(y)

    grad_output_wp = wp.from_torch(grad_output.detach(), return_ctype=True)
    x_wp = wp.from_torch(x.detach(), return_ctype=True)
    y_wp = wp.from_torch(y.detach(), return_ctype=True)
    grad_x_wp = wp.from_torch(grad_x.detach(), return_ctype=True)
    grad_y_wp = wp.from_torch(grad_y.detach(), return_ctype=True)
    tensor_matmul_o3_3x3_bwd = get_module("tensor_matmul_o3_3x3_bwd", [str(x.dtype)])
    wp.launch(
        tensor_matmul_o3_3x3_bwd,
        dim=(x.shape[0], x.shape[-1]),
        stream=stream,
        device=device,
        inputs=(x_wp, y_wp, grad_output_wp, grad_x_wp, grad_y_wp),
    )

    return [grad_x, grad_y]


@torch.library.register_fake("tensornet::tensor_matmul_o3_3x3_bwd_primitive")
def _(grad_output: list[Tensor], x: Tensor, y: Tensor) -> list[Tensor]:
    return [torch.empty_like(x), torch.empty_like(y)]


@torch.library.custom_op(
    "tensornet::tensor_matmul_o3_3x3_bwd_bwd_primitive",
    mutates_args=(),
    device_types=["cpu", "cuda"],
)
def _(
    grad_output: Tensor, grad_grad_x: Tensor, grad_grad_y: Tensor, x: Tensor, y: Tensor
) -> list[Tensor]:
    if x.shape[1] != 3 or x.shape[2] != 3 or y.shape[1] != 3 or y.shape[2] != 3:
        raise ValueError("x and y must be 3x3 matrices")
    if x.ndim != 4 or y.ndim != 4:
        raise ValueError("x and y must be 4D tensors")

    stream = get_stream(grad_output.device)
    device = wp.device_from_torch(grad_output.device)
    grad_x = torch.empty_like(grad_output)
    grad_y = torch.empty_like(grad_output)

    grad_grad_output = torch.empty_like(grad_output)

    grad_grad_x_wp = wp.from_torch(grad_grad_x.detach(), return_ctype=True)
    grad_grad_y_wp = wp.from_torch(grad_grad_y.detach(), return_ctype=True)
    grad_output_wp = wp.from_torch(grad_output.detach(), return_ctype=True)
    x_wp = wp.from_torch(x.detach(), return_ctype=True)
    y_wp = wp.from_torch(y.detach(), return_ctype=True)

    grad_x_wp = wp.from_torch(grad_x.detach(), return_ctype=True)
    grad_y_wp = wp.from_torch(grad_y.detach(), return_ctype=True)
    grad_grad_output_wp = wp.from_torch(grad_grad_output.detach(), return_ctype=True)

    tensor_matmul_o3_3x3_bwd_bwd = get_module(
        "tensor_matmul_o3_3x3_bwd_bwd", [str(grad_output.dtype)]
    )
    wp.launch(
        tensor_matmul_o3_3x3_bwd_bwd,
        dim=(grad_output.shape[0], grad_output.shape[-1]),
        stream=stream,
        device=device,
        inputs=(
            x_wp,
            y_wp,
            grad_grad_x_wp,
            grad_grad_y_wp,
            grad_output_wp,
            grad_x_wp,
            grad_y_wp,
            grad_grad_output_wp,
        ),
    )

    return [grad_grad_output, grad_x, grad_y]


@torch.library.register_fake("tensornet::tensor_matmul_o3_3x3_bwd_bwd_primitive")
def _(
    grad_output: Tensor, grad_grad_x: Tensor, grad_grad_y: Tensor, x: Tensor, y: Tensor
) -> list[Tensor]:
    return [
        torch.empty_like(grad_output),
        torch.empty_like(grad_output),
        torch.empty_like(grad_output),
    ]


def tensor_matmul_o3_3x3_setup_fwd_context(ctx, inputs, output):
    (x, y) = inputs
    ctx.save_for_backward(x, y)


def tensor_matmul_o3_3x3_setup_bwd_context(ctx, inputs, output):
    (grad_output, x, y) = inputs
    ctx.save_for_backward(grad_output, x, y)


@torch.compiler.allow_in_graph
def tensor_matmul_o3_3x3_fwd(*args):
    return torch.ops.tensornet.tensor_matmul_o3_3x3_fwd_primitive(*args)


@torch.compiler.allow_in_graph
def tensor_matmul_o3_3x3_bwd(ctx, grad_output):
    x, y = ctx.saved_tensors
    dx, dy = torch.ops.tensornet.tensor_matmul_o3_3x3_bwd_primitive(grad_output, x, y)
    return dx, dy


@torch.compiler.allow_in_graph
def tensor_matmul_o3_3x3_bwd_bwd(ctx, *grad_outputs):
    grad_grad_x = grad_outputs[0][0]
    grad_grad_y = grad_outputs[0][1]

    grad_output_saved, x, y = ctx.saved_tensors

    if grad_grad_x is None:
        grad_grad_x = torch.zeros_like(x)
    if grad_grad_y is None:
        grad_grad_y = torch.zeros_like(y)

    outputs = torch.ops.tensornet.tensor_matmul_o3_3x3_bwd_bwd_primitive(
        grad_output_saved, grad_grad_x, grad_grad_y, x, y
    )
    return outputs[0], outputs[1], outputs[2]


torch.library.register_autograd(
    "tensornet::tensor_matmul_o3_3x3_fwd_primitive",
    tensor_matmul_o3_3x3_bwd,
    setup_context=tensor_matmul_o3_3x3_setup_fwd_context,
)

torch.library.register_autograd(
    "tensornet::tensor_matmul_o3_3x3_bwd_primitive",
    tensor_matmul_o3_3x3_bwd_bwd,
    setup_context=tensor_matmul_o3_3x3_setup_bwd_context,
)


def fn_tensor_matmul_o3_3x3(x: Tensor, y: Tensor) -> Tensor:
    z = torch.ops.tensornet.tensor_matmul_o3_3x3_fwd_primitive(x, y)
    return z



# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

from os.path import join
from tqdm import tqdm
import torch
from torch.utils.data import Subset
from torch_geometric.loader import DataLoader
from lightning import LightningDataModule
from lightning_utilities.core.rank_zero import rank_zero_warn
from torchmdnet import datasets
from torchmdnet.utils import make_splits, MissingEnergyException
from torchmdnet.models.utils import scatter
import warnings


class DataModule(LightningDataModule):
    """A LightningDataModule for loading datasets from the torchmdnet.datasets module.

    Args:
        hparams (dict): A dictionary containing the hyperparameters of the
            dataset. See the documentation of the torchmdnet.datasets module
            for details.
        dataset (torch_geometric.data.Dataset): A dataset to use instead of
            loading a new one from the torchmdnet.datasets module.
    """

    def __init__(self, hparams, dataset=None):
        super(DataModule, self).__init__()
        self.save_hyperparameters(hparams)
        self._mean, self._std = None, None
        self._saved_dataloaders = dict()
        self.dataset = dataset

    def setup(self, stage):
        if self.dataset is None:
            if self.hparams["dataset"] == "Custom":
                self.dataset = datasets.Custom(
                    self.hparams["coord_files"],
                    self.hparams["embed_files"],
                    self.hparams["energy_files"],
                    self.hparams["force_files"],
                    self.hparams["dataset_preload_limit"],
                )
            else:
                dataset_arg = {}
                if self.hparams["dataset_arg"] is not None:
                    dataset_arg = self.hparams["dataset_arg"]
                if self.hparams["dataset"] == "HDF5":
                    dataset_arg["dataset_preload_limit"] = self.hparams[
                        "dataset_preload_limit"
                    ]
                self.dataset = getattr(datasets, self.hparams["dataset"])(
                    self.hparams["dataset_root"], **dataset_arg
                )

        self.idx_train, self.idx_val, self.idx_test = make_splits(
            len(self.dataset),
            self.hparams["train_size"],
            self.hparams["val_size"],
            self.hparams["test_size"],
            self.hparams["seed"],
            join(self.hparams["log_dir"], "splits.npz"),
            self.hparams["splits"],
        )
        print(
            f"train {len(self.idx_train)}, val {len(self.idx_val)}, test {len(self.idx_test)}"
        )

        self.train_dataset = Subset(self.dataset, self.idx_train)
        self.val_dataset = Subset(self.dataset, self.idx_val)
        self.test_dataset = Subset(self.dataset, self.idx_test)

        if self.hparams["standardize"]:
            # Mark as deprecated
            warnings.warn(
                "The standardize option is deprecated and will be removed in the future. ",
                DeprecationWarning,
            )
            self._standardize()

    def train_dataloader(self):
        return self._get_dataloader(self.train_dataset, "train")

    def val_dataloader(self):
        loaders = [self._get_dataloader(self.val_dataset, "val")]
        # To allow to report the performance on the testing dataset during training
        # we send the trainer two dataloaders every few steps and modify the
        # validation step to understand the second dataloader as test data.
        if self._is_test_during_training_epoch():
            loaders.append(self._get_dataloader(self.test_dataset, "test"))
        return loaders

    def test_dataloader(self):
        return self._get_dataloader(self.test_dataset, "test")

    @property
    def atomref(self):
        """Returns the atomref of the dataset if it has one, otherwise None."""
        if hasattr(self.dataset, "get_atomref"):
            return self.dataset.get_atomref()
        return None

    @property
    def mean(self):
        """Returns the mean of the dataset if it has one, otherwise None."""
        return self._mean

    @property
    def std(self):
        """Returns the standard deviation of the dataset if it has one, otherwise None."""
        return self._std

    def _is_test_during_training_epoch(self):
        return (
            len(self.test_dataset) > 0
            and self.hparams["test_interval"] > 0
            and self.trainer.current_epoch > 0
            and self.trainer.current_epoch % self.hparams["test_interval"] == 0
        )

    def _get_dataloader(self, dataset, stage, store_dataloader=True):
        if stage in self._saved_dataloaders and store_dataloader:
            return self._saved_dataloaders[stage]

        if stage == "train":
            batch_size = self.hparams["batch_size"]
        elif stage in ["val", "test"]:
            batch_size = self.hparams["inference_batch_size"]

        shuffle = stage == "train"
        dl = DataLoader(
            dataset=dataset,
            batch_size=batch_size,
            num_workers=self.hparams["num_workers"],
            persistent_workers=False,
            pin_memory=True,
            shuffle=shuffle,
        )

        if store_dataloader:
            self._saved_dataloaders[stage] = dl
        return dl

    def _standardize(self):
        def get_energy(batch, atomref):
            if "y" not in batch or batch.y is None:
                raise MissingEnergyException()

            if atomref is None:
                return batch.y.clone()

            # remove atomref energies from the target energy
            atomref_energy = scatter(atomref[batch.z], batch.batch, dim=0)
            return (batch.y.squeeze() - atomref_energy.squeeze()).clone()

        data = tqdm(
            self._get_dataloader(self.train_dataset, "val", store_dataloader=False),
            desc="computing mean and std",
        )
        try:
            # only remove atomref energies if the atomref prior is used
            atomref = self.atomref if self.hparams["prior_model"] == "Atomref" else None
            # extract energies from the data
            ys = torch.cat([get_energy(batch, atomref) for batch in data])
        except MissingEnergyException:
            rank_zero_warn(
                "Standardize is true but failed to compute dataset mean and "
                "standard deviation. Maybe the dataset only contains forces."
            )
            return

        # compute mean and standard deviation
        self._mean = ys.mean(dim=0)
        self._std = ys.std(dim=0)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

import torch
import warp as wp
from torch import Tensor

from torchmdnet.extensions.warp_kernels import get_module, get_stream


@torch.library.custom_op(
    "tensornet::radial_message_passing_fwd_primitive",
    mutates_args=(),
    device_types=["cpu", "cuda"],
)
def _(
    edge_vec_norm: Tensor, edge_attr: Tensor, row_data: Tensor, row_indptr: Tensor
) -> list[Tensor]:
    num_atoms = row_indptr.shape[0] - 1
    stream = get_stream(edge_vec_norm.device)
    device = wp.device_from_torch(edge_vec_norm.device)
    output_I = torch.zeros(
        (num_atoms, 1, edge_attr.shape[-1]),
        dtype=edge_vec_norm.dtype,
        device=edge_vec_norm.device,
    )
    output_A = torch.zeros(
        (num_atoms, 3, edge_attr.shape[-1]),
        dtype=edge_vec_norm.dtype,
        device=edge_vec_norm.device,
    )
    output_S = torch.zeros(
        (num_atoms, 5, edge_attr.shape[-1]),
        dtype=edge_vec_norm.dtype,
        device=edge_vec_norm.device,
    )

    output_I_wp = wp.from_torch(output_I.detach(), return_ctype=True)
    output_A_wp = wp.from_torch(output_A.detach(), return_ctype=True)
    output_S_wp = wp.from_torch(output_S.detach(), return_ctype=True)

    edge_vec_norm_wp = wp.from_torch(edge_vec_norm.detach(), return_ctype=True)
    edge_attr_wp = wp.from_torch(edge_attr.detach(), return_ctype=True)

    row_data_wp = wp.from_torch(row_data.detach(), return_ctype=True)
    row_indptr_wp = wp.from_torch(row_indptr.detach(), return_ctype=True)

    message_passing_fwd = get_module(
        "radial_message_passing_fwd", [str(edge_vec_norm.dtype)]
    )
    wp.launch(
        message_passing_fwd,
        dim=(num_atoms, edge_attr.shape[-1]),
        stream=stream,
        device=device,
        inputs=(
            edge_vec_norm_wp,
            edge_attr_wp,
            row_data_wp,
            row_indptr_wp,
            output_I_wp,
            output_A_wp,
            output_S_wp,
        ),
    )

    return [output_I, output_A, output_S]


@torch.library.register_fake("tensornet::radial_message_passing_fwd_primitive")
def _(
    edge_vec_norm: Tensor, edge_attr: Tensor, row_data: Tensor, row_indptr: Tensor
) -> list[Tensor]:
    num_atoms = row_indptr.shape[0] - 1
    return [
        torch.empty(
            (num_atoms, 1, edge_attr.shape[-1]),
            dtype=edge_vec_norm.dtype,
            device=edge_vec_norm.device,
        ),
        torch.empty(
            (num_atoms, 3, edge_attr.shape[-1]),
            dtype=edge_vec_norm.dtype,
            device=edge_vec_norm.device,
        ),
        torch.empty(
            (num_atoms, 5, edge_attr.shape[-1]),
            dtype=edge_vec_norm.dtype,
            device=edge_vec_norm.device,
        ),
    ]


@torch.library.custom_op(
    "tensornet::radial_message_passing_bwd_primitive",
    mutates_args=(),
    device_types=["cpu", "cuda"],
)
def _(
    grad_output_I: Tensor,
    grad_output_A: Tensor,
    grad_output_S: Tensor,
    edge_vec_norm: Tensor,
    edge_attr: Tensor,
    row_data: Tensor,
    row_indptr: Tensor,
) -> list[Tensor]:
    num_atoms = row_indptr.shape[0] - 1
    stream = get_stream(grad_output_I.device)
    device = wp.device_from_torch(grad_output_I.device)

    grad_output_I_wp = wp.from_torch(grad_output_I.detach(), return_ctype=True)
    grad_output_A_wp = wp.from_torch(grad_output_A.detach(), return_ctype=True)
    grad_output_S_wp = wp.from_torch(grad_output_S.detach(), return_ctype=True)

    grad_edge_vec_norm = torch.zeros_like(edge_vec_norm)
    grad_edge_vec_norm_wp = wp.from_torch(
        grad_edge_vec_norm.detach(), return_ctype=True
    )

    grad_edge_attr = torch.zeros_like(edge_attr)
    grad_edge_attr_wp = wp.from_torch(grad_edge_attr.detach(), return_ctype=True)

    edge_vec_norm_wp = wp.from_torch(edge_vec_norm.detach(), return_ctype=True)
    edge_attr_wp = wp.from_torch(edge_attr.detach(), return_ctype=True)

    row_data_wp = wp.from_torch(row_data.detach(), return_ctype=True)
    row_indptr_wp = wp.from_torch(row_indptr.detach(), return_ctype=True)

    message_passing_bwd = get_module(
        "radial_message_passing_bwd", [str(edge_vec_norm.dtype)]
    )
    wp.launch(
        message_passing_bwd,
        dim=(num_atoms, edge_attr.shape[-1]),
        stream=stream,
        device=device,
        inputs=(
            edge_vec_norm_wp,
            edge_attr_wp,
            row_data_wp,
            row_indptr_wp,
            grad_output_I_wp,
            grad_output_A_wp,
            grad_output_S_wp,
            grad_edge_vec_norm_wp,
            grad_edge_attr_wp,
        ),
    )

    return [grad_edge_vec_norm, grad_edge_attr]


@torch.library.register_fake("tensornet::radial_message_passing_bwd_primitive")
def _(
    grad_output_I: Tensor,
    grad_output_A: Tensor,
    grad_output_S: Tensor,
    edge_vec_norm: Tensor,
    edge_attr: Tensor,
    row_data: Tensor,
    row_indptr: Tensor,
) -> list[Tensor]:
    return [torch.empty_like(edge_vec_norm), torch.empty_like(edge_attr)]


@torch.library.custom_op(
    "tensornet::radial_message_passing_bwd_bwd_primitive",
    mutates_args=(),
    device_types=["cpu", "cuda"],
)
def _(
    grad_output_I: Tensor,
    grad_output_A: Tensor,
    grad_output_S: Tensor,
    grad_grad_edge_vec_norm: Tensor,
    grad_grad_edge_attr: Tensor,
    edge_vec_norm: Tensor,
    edge_attr: Tensor,
    row_data: Tensor,
    row_indptr: Tensor,
) -> list[Tensor]:
    num_atoms = row_indptr.shape[0] - 1
    stream = get_stream(grad_output_I.device)
    device = wp.device_from_torch(grad_output_I.device)

    edge_vec_norm_wp = wp.from_torch(edge_vec_norm.detach(), return_ctype=True)
    edge_attr_wp = wp.from_torch(edge_attr.detach(), return_ctype=True)

    row_data_wp = wp.from_torch(row_data.detach(), return_ctype=True)
    row_indptr_wp = wp.from_torch(row_indptr.detach(), return_ctype=True)

    grad_grad_edge_vec_norm_wp = wp.from_torch(
        grad_grad_edge_vec_norm.detach(), return_ctype=True
    )
    grad_grad_edge_attr_wp = wp.from_torch(
        grad_grad_edge_attr.detach(), return_ctype=True
    )

    grad_output_I_wp = wp.from_torch(grad_output_I.detach(), return_ctype=True)
    grad_output_A_wp = wp.from_torch(grad_output_A.detach(), return_ctype=True)
    grad_output_S_wp = wp.from_torch(grad_output_S.detach(), return_ctype=True)
    dgrad_output_I = torch.zeros_like(grad_output_I)
    dgrad_output_A = torch.zeros_like(grad_output_A)
    dgrad_output_S = torch.zeros_like(grad_output_S)
    dgrad_output_I_wp = wp.from_torch(dgrad_output_I.detach(), return_ctype=True)
    dgrad_output_A_wp = wp.from_torch(dgrad_output_A.detach(), return_ctype=True)
    dgrad_output_S_wp = wp.from_torch(dgrad_output_S.detach(), return_ctype=True)

    dgrad_grad_edge_vec_norm = torch.zeros_like(grad_grad_edge_vec_norm)
    dgrad_grad_edge_vec_norm_wp = wp.from_torch(
        dgrad_grad_edge_vec_norm.detach(), return_ctype=True
    )

    dgrad_grad_edge_attr = torch.zeros_like(grad_grad_edge_attr)
    dgrad_grad_edge_attr_wp = wp.from_torch(
        dgrad_grad_edge_attr.detach(), return_ctype=True
    )

    message_passing_bwd_bwd = get_module(
        "radial_message_passing_bwd_bwd", [str(edge_vec_norm.dtype)]
    )
    wp.launch(
        message_passing_bwd_bwd,
        dim=(num_atoms, edge_attr.shape[-1]),
        stream=stream,
        device=device,
        inputs=(
            edge_vec_norm_wp,
            edge_attr_wp,
            grad_grad_edge_vec_norm_wp,
            grad_grad_edge_attr_wp,
            grad_output_I_wp,
            grad_output_A_wp,
            grad_output_S_wp,
            row_data_wp,
            row_indptr_wp,
            dgrad_grad_edge_vec_norm_wp,
            dgrad_grad_edge_attr_wp,
            dgrad_output_I_wp,
            dgrad_output_A_wp,
            dgrad_output_S_wp,
        ),
    )

    return [
        dgrad_output_I,
        dgrad_output_A,
        dgrad_output_S,
        dgrad_grad_edge_vec_norm,
        dgrad_grad_edge_attr,
    ]


@torch.library.register_fake("tensornet::radial_message_passing_bwd_bwd_primitive")
def _(
    grad_output_I: Tensor,
    grad_output_A: Tensor,
    grad_output_S: Tensor,
    grad_grad_edge_vec_norm: Tensor,
    grad_grad_edge_attr: Tensor,
    edge_vec_norm: Tensor,
    edge_attr: Tensor,
    row_data: Tensor,
    row_indptr: Tensor,
) -> list[Tensor]:
    return [
        torch.empty_like(grad_output_I),
        torch.empty_like(grad_output_A),
        torch.empty_like(grad_output_S),
        torch.empty_like(grad_grad_edge_vec_norm),
        torch.empty_like(grad_grad_edge_attr),
    ]


def radial_message_passing_setup_fwd_context(ctx, inputs, output):
    (edge_vec_norm, edge_attr, row_data, row_indptr) = inputs
    ctx.save_for_backward(edge_vec_norm, edge_attr, row_data, row_indptr)


def radial_message_passing_setup_bwd_context(ctx, inputs, output):
    (
        grad_output_I,
        grad_output_A,
        grad_output_S,
        edge_vec_norm,
        edge_attr,
        row_data,
        row_indptr,
    ) = inputs
    ctx.save_for_backward(
        grad_output_I,
        grad_output_A,
        grad_output_S,
        edge_vec_norm,
        edge_attr,
        row_data,
        row_indptr,
    )


@torch.compiler.allow_in_graph
def radial_message_passing_fwd(*args):
    return torch.ops.tensornet.radial_message_passing_fwd_primitive(*args)


@torch.compiler.allow_in_graph
def radial_message_passing_bwd(ctx, grad_outputs):
    edge_vec_norm, edge_attr, row_data, row_indptr = ctx.saved_tensors

    result = torch.ops.tensornet.radial_message_passing_bwd_primitive(
        grad_outputs[0],
        grad_outputs[1],
        grad_outputs[2],
        edge_vec_norm,
        edge_attr,
        row_data,
        row_indptr,
    )

    grad_edge_vec_norm, grad_edge_attr = result

    return grad_edge_vec_norm, grad_edge_attr, None, None


@torch.compiler.allow_in_graph
def radial_message_passing_bwd_bwd(ctx, *grad_outputs):
    grad_grad_edge_vec_norm, grad_grad_edge_attr = grad_outputs[0]
    (
        grad_output_I,
        grad_output_A,
        grad_output_S,
        edge_vec_norm,
        edge_attr,
        row_data,
        row_indptr,
    ) = ctx.saved_tensors

    result = torch.ops.tensornet.radial_message_passing_bwd_bwd_primitive(
        grad_output_I,
        grad_output_A,
        grad_output_S,
        grad_grad_edge_vec_norm,
        grad_grad_edge_attr,
        edge_vec_norm,
        edge_attr,
        row_data,
        row_indptr,
    )

    (
        dgrad_output_I,
        dgrad_output_A,
        dgrad_output_S,
        dgrad_grad_edge_vec_norm,
        dgrad_grad_edge_attr,
    ) = result

    return (
        dgrad_output_I,
        dgrad_output_A,
        dgrad_output_S,
        dgrad_grad_edge_vec_norm,
        dgrad_grad_edge_attr,
        None,
        None,
    )


torch.library.register_autograd(
    "tensornet::radial_message_passing_fwd_primitive",
    radial_message_passing_bwd,
    setup_context=radial_message_passing_setup_fwd_context,
)

torch.library.register_autograd(
    "tensornet::radial_message_passing_bwd_primitive",
    radial_message_passing_bwd_bwd,
    setup_context=radial_message_passing_setup_bwd_context,
)


def fn_radial_message_passing(
    edge_vec_norm: Tensor, edge_attr: Tensor, row_data: Tensor, row_indptr: Tensor
) -> list[Tensor]:
    return torch.ops.tensornet.radial_message_passing_fwd_primitive(
        edge_vec_norm, edge_attr, row_data, row_indptr
    )



# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

import math
from typing import Optional, Tuple
import torch
from torch import nn, Tensor
import torch.nn.functional as F
from torchmdnet.extensions.ops import get_neighbor_pairs_kernel
import warnings


def visualize_basis(basis_type, num_rbf=50, cutoff_lower=0, cutoff_upper=5):
    """
    Function for quickly visualizing a specific basis. This is useful for inspecting
    the distance coverage of basis functions for non-default lower and upper cutoffs.

    Args:
        basis_type (str): Specifies the type of basis functions used. Can be one of
            ['gauss',expnorm']
        num_rbf (int, optional): The number of basis functions.
            (default: :obj:`50`)
        cutoff_lower (float, optional): The lower cutoff of the basis.
            (default: :obj:`0`)
        cutoff_upper (float, optional): The upper cutoff of the basis.
            (default: :obj:`5`)
    """
    import matplotlib.pyplot as plt

    distances = torch.linspace(cutoff_lower - 1, cutoff_upper + 1, 1000)
    basis_kwargs = {
        "num_rbf": num_rbf,
        "cutoff_lower": cutoff_lower,
        "cutoff_upper": cutoff_upper,
    }
    basis_expansion = rbf_class_mapping[basis_type](**basis_kwargs)
    expanded_distances = basis_expansion(distances)

    for i in range(expanded_distances.shape[-1]):
        plt.plot(distances.numpy(), expanded_distances[:, i].detach().numpy())
    plt.show()


class NeighborEmbedding(nn.Module):
    def __init__(
        self,
        hidden_channels,
        num_rbf,
        cutoff_lower,
        cutoff_upper,
        max_z=100,
        dtype=torch.float32,
    ):
        """
        The ET architecture assigns two  learned vectors to each atom type
        zi. One  is used to  encode information  specific to an  atom, the
        other (this  class) takes  the role  of a  neighborhood embedding.
        The neighborhood embedding, which is  an embedding of the types of
        neighboring atoms, is multiplied by a distance filter.


        This embedding allows  the network to store  information about the
        interaction of atom pairs.

        See eq. 3 in https://arxiv.org/pdf/2202.02541.pdf for more details.
        """
        super(NeighborEmbedding, self).__init__()
        self.embedding = nn.Embedding(max_z, hidden_channels, dtype=dtype)
        self.distance_proj = nn.Linear(num_rbf, hidden_channels, dtype=dtype)
        self.combine = nn.Linear(hidden_channels * 2, hidden_channels, dtype=dtype)
        self.cutoff = CosineCutoff(cutoff_lower, cutoff_upper)

        self.reset_parameters()

    def reset_parameters(self):
        self.embedding.reset_parameters()
        nn.init.xavier_uniform_(self.distance_proj.weight)
        nn.init.xavier_uniform_(self.combine.weight)
        self.distance_proj.bias.data.fill_(0)
        self.combine.bias.data.fill_(0)

    def forward(
        self,
        z: Tensor,
        x: Tensor,
        edge_index: Tensor,
        edge_weight: Tensor,
        edge_attr: Tensor,
    ) -> Tensor:
        """
        Args:
            z (Tensor): Atomic numbers of shape :obj:`[num_nodes]`
            x (Tensor): Node feature matrix (atom positions) of shape :obj:`[num_nodes, 3]`
            edge_index (Tensor): Graph connectivity (list of neighbor pairs) with shape :obj:`[2, num_edges]`
            edge_weight (Tensor): Edge weight vector of shape :obj:`[num_edges]`
            edge_attr (Tensor): Edge attribute matrix of shape :obj:`[num_edges, 3]`
        Returns:
            x_neighbors (Tensor): The embedding of the neighbors of each atom of shape :obj:`[num_nodes, hidden_channels]`
        """
        # remove self loops
        mask = edge_index[0] != edge_index[1]
        if not mask.all():
            edge_index = edge_index[:, mask]
            edge_weight = edge_weight[mask]
            edge_attr = edge_attr[mask]

        C = self.cutoff(edge_weight)
        W = self.distance_proj(edge_attr) * C.view(-1, 1)

        x_neighbors = self.embedding(z)
        msg = W * x_neighbors.index_select(0, edge_index[1])
        x_neighbors = torch.zeros(
            z.shape[0], x.shape[1], dtype=x.dtype, device=x.device
        ).index_add(0, edge_index[0], msg)
        x_neighbors = self.combine(torch.cat([x, x_neighbors], dim=1))
        return x_neighbors


class OptimizedDistance(torch.nn.Module):
    """Compute the neighbor list for a given cutoff.

    This operation can be placed inside a CUDA graph in some cases.
    In particular, resize_to_fit must be False.

    Note that this module returns neighbors such that :math:`r_{ij} \\ge \\text{cutoff_lower}\\quad\\text{and}\\quad r_{ij} < \\text{cutoff_upper}`.

    This function optionally supports periodic boundary conditions with
    arbitrary triclinic boxes.  The box vectors `a`, `b`, and `c` must satisfy
    certain requirements:

    .. code:: python

       a[1] = a[2] = b[2] = 0
       a[0] >= 2*cutoff, b[1] >= 2*cutoff, c[2] >= 2*cutoff
       a[0] >= 2*b[0]
       a[0] >= 2*c[0]
       b[1] >= 2*c[1]

    These requirements correspond to a particular rotation of the system and
    reduced form of the vectors, as well as the requirement that the cutoff be
    no larger than half the box width.

    Parameters
    ----------
    cutoff_lower : float
        Lower cutoff for the neighbor list.
    cutoff_upper : float
        Upper cutoff for the neighbor list.
    max_num_pairs : int
        Maximum number of pairs to store, if the number of pairs found is less than this, the list is padded with (-1,-1) pairs up to max_num_pairs unless resize_to_fit is True, in which case the list is resized to the actual number of pairs found.
        If the number of pairs found is larger than this, an asynchronous device-side error is raised.
        If negative, it is interpreted as (minus) the maximum number of neighbors per atom.
    strategy : str
        Strategy to use for computing the neighbor list. Can be one of :code:`["brute", "cell"]`.

        1. *Shared*: An O(N^2) algorithm that leverages CUDA shared memory, best for large number of particles.
        2. *Brute*: A brute force O(N^2) algorithm, best for small number of particles.
        3. *Cell*:  A cell list algorithm, best for large number of particles, low cutoffs and low batch size.
    box : torch.Tensor, optional
        The vectors defining the periodic box.  This must have shape `(3, 3)` or `(max(batch)+1, 3, 3)` if a ox per sample is desired.
        where `box_vectors[0] = a`, `box_vectors[1] = b`, and `box_vectors[2] = c`.
        If this is omitted, periodic boundary conditions are not applied.
    loop : bool, optional
        Whether to include self-interactions.
        Default: False
    include_transpose : bool, optional
        Whether to include the transpose of the neighbor list.
        Default: True
    resize_to_fit : bool, optional
        Whether to resize the neighbor list to the actual number of pairs found. When False, the list is padded with (-1,-1) pairs up to max_num_pairs
        Default: True
        If this is True the operation is not CUDA graph compatible.
    return_vecs : bool, optional
        Whether to return the distance vectors.
        Default: False
    long_edge_index : bool, optional
        Whether to return edge_index as int64, otherwise int32.
        Default: True
    """

    def __init__(
        self,
        cutoff_lower=0.0,
        cutoff_upper=5.0,
        max_num_pairs=-32,
        return_vecs=False,
        loop=False,
        strategy="brute",
        include_transpose=True,
        resize_to_fit=True,
        box=None,
        long_edge_index=True,
    ):
        super(OptimizedDistance, self).__init__()
        self.cutoff_upper = cutoff_upper
        self.cutoff_lower = cutoff_lower
        self.max_num_pairs = max_num_pairs
        self.strategy = strategy
        self.loop = loop
        self.return_vecs = return_vecs
        self.include_transpose = include_transpose
        self.resize_to_fit = resize_to_fit
        self.use_periodic = True
        self.num_cells = 0

        # Use register_buffer for box to make it export-compatible and handle device movement
        if box is None:
            self.use_periodic = False
            if strategy == "cell":
                # Default the box to 3 times the cutoff, really inefficient for the cell list
                lbox = cutoff_upper * 3.0
                box = torch.tensor(
                    [[lbox, 0, 0], [0, lbox, 0], [0, 0, lbox]], device="cpu"
                )
            else:
                # Use a placeholder box instead of empty (0,0) to avoid shape issues in torch.export
                # This won't be used when use_periodic=False, but prevents export tracing issues
                box = torch.zeros((3, 3), device="cpu")

        # Register box as a buffer so it moves with the module and is export-compatible
        self.register_buffer("box", box, persistent=True)

        if self.strategy == "cell":
            from torchmdnet.extensions.triton_cell import _get_cell_dimensions

            cell_dims = _get_cell_dimensions(
                self.box[0, 0], self.box[1, 1], self.box[2, 2], cutoff_upper
            )
            self.num_cells = int(cell_dims.prod())
            if self.num_cells > 1024**3:
                raise RuntimeError(
                    f"Too many cells: {self.num_cells}. Maximum is 1024^3. "
                    f"Reduce box size or increase cutoff."
                )

        self.long_edge_index = long_edge_index

    def forward(
        self, pos: Tensor, batch: Optional[Tensor] = None, box: Optional[Tensor] = None
    ) -> Tuple[Tensor, Tensor, Optional[Tensor]]:
        """
        Compute the neighbor list for a given cutoff.

        Parameters
        ----------
        pos : torch.Tensor
            A tensor with shape (N, 3) representing the positions.
        batch : torch.Tensor, optional
            A tensor with shape (N,). Defaults to None.
        box : torch.Tensor, optional
            The vectors defining the periodic box.  This must have shape `(3, 3)` or `(max(batch)+1, 3, 3)`,
        Returns
        -------
        edge_index : torch.Tensor
            List of neighbors for each atom in the batch.
            Shape is (2, num_found_pairs) or (2, max_num_pairs).
        edge_weight : torch.Tensor
            List of distances for each atom in the batch.
            Shape is (num_found_pairs,) or (max_num_pairs,).
        edge_vec : torch.Tensor, optional
            List of distance vectors for each atom in the batch.
            Shape is (num_found_pairs, 3) or (max_num_pairs, 3).

        Notes
        -----
        If `resize_to_fit` is True, the tensors will be trimmed to the actual number of pairs found.
        Otherwise, the tensors will have size `max_num_pairs`, with neighbor pairs (-1, -1) at the end.
        """
        use_periodic = self.use_periodic
        if not use_periodic:
            use_periodic = box is not None

        # Use the registered buffer box if no box is provided
        # The buffer is automatically moved to the correct device with the module
        if box is None:
            box = self.box

        assert box is not None, "Box must be provided"

        # Ensure box has correct dtype (device is already correct if using self.box)
        if box.dtype != pos.dtype:
            box = box.to(dtype=pos.dtype)

        max_pairs = self.max_num_pairs
        if self.max_num_pairs < 0:
            max_pairs = -self.max_num_pairs * pos.shape[0]
        if batch is None:
            batch = torch.zeros(pos.shape[0], dtype=torch.long, device=pos.device)
        edge_index, edge_vec, edge_weight, num_pairs = get_neighbor_pairs_kernel(
            strategy=self.strategy,
            positions=pos,
            batch=batch,
            max_num_pairs=max_pairs,
            cutoff_lower=self.cutoff_lower,
            cutoff_upper=self.cutoff_upper,
            loop=self.loop,
            include_transpose=self.include_transpose,
            box_vectors=box,
            use_periodic=use_periodic,
            num_cells=self.num_cells,
        )
        torch._assert_async(
            (num_pairs[0] <= max_pairs),
            "Found num_pairs > max_num_pairs, please increase max_num_pairs",
        )

        # Remove (-1,-1)  pairs
        if self.resize_to_fit:
            mask = edge_index[0] != -1
            edge_index = edge_index[:, mask]
            edge_weight = edge_weight[mask]
            edge_vec = edge_vec[mask, :]
        if self.long_edge_index:
            edge_index = edge_index.to(torch.long)
        if self.return_vecs:
            return edge_index, edge_weight, edge_vec
        else:
            return edge_index, edge_weight, None


class GaussianSmearing(nn.Module):
    def __init__(
        self,
        cutoff_lower=0.0,
        cutoff_upper=5.0,
        num_rbf=50,
        trainable=True,
        dtype=torch.float32,
    ):
        super(GaussianSmearing, self).__init__()
        self.cutoff_lower = cutoff_lower
        self.cutoff_upper = cutoff_upper
        self.num_rbf = num_rbf
        self.trainable = trainable
        self.dtype = dtype
        offset, coeff = self._initial_params()
        if trainable:
            self.register_parameter("coeff", nn.Parameter(coeff))
            self.register_parameter("offset", nn.Parameter(offset))
        else:
            self.register_buffer("coeff", coeff)
            self.register_buffer("offset", offset)

    def _initial_params(self):
        offset = torch.linspace(
            self.cutoff_lower, self.cutoff_upper, self.num_rbf, dtype=self.dtype
        )
        coeff = -0.5 / (offset[1] - offset[0]) ** 2
        return offset, coeff

    def reset_parameters(self):
        offset, coeff = self._initial_params()
        self.offset.data.copy_(offset)
        self.coeff.data.copy_(coeff)

    def forward(self, dist: Tensor) -> Tensor:
        dist = dist.unsqueeze(-1) - self.offset
        return torch.exp(self.coeff * torch.pow(dist, 2))


class ExpNormalSmearing(nn.Module):
    def __init__(
        self,
        cutoff_lower=0.0,
        cutoff_upper=5.0,
        num_rbf=50,
        trainable=True,
        dtype=torch.float32,
    ):
        super(ExpNormalSmearing, self).__init__()
        self.cutoff_lower = cutoff_lower
        self.cutoff_upper = cutoff_upper
        self.num_rbf = num_rbf
        self.trainable = trainable
        self.dtype = dtype
        self.cutoff_fn = CosineCutoff(0, cutoff_upper)
        self.alpha = 5.0 / (cutoff_upper - cutoff_lower)

        means, betas = self._initial_params()
        if trainable:
            self.register_parameter("means", nn.Parameter(means))
            self.register_parameter("betas", nn.Parameter(betas))
        else:
            self.register_buffer("means", means)
            self.register_buffer("betas", betas)

    def _initial_params(self):
        # initialize means and betas according to the default values in PhysNet
        # https://pubs.acs.org/doi/10.1021/acs.jctc.9b00181
        start_value = torch.exp(
            torch.scalar_tensor(
                -self.cutoff_upper + self.cutoff_lower, dtype=self.dtype
            )
        )
        means = torch.linspace(start_value, 1, self.num_rbf, dtype=self.dtype)
        betas = torch.tensor(
            [(2 / self.num_rbf * (1 - start_value)) ** -2] * self.num_rbf,
            dtype=self.dtype,
        )
        return means, betas

    def reset_parameters(self):
        means, betas = self._initial_params()
        self.means.data.copy_(means)
        self.betas.data.copy_(betas)

    def forward(self, dist):
        dist = dist.unsqueeze(-1)
        return self.cutoff_fn(dist) * torch.exp(
            -self.betas
            * (torch.exp(self.alpha * (-dist + self.cutoff_lower)) - self.means) ** 2
        )


class GLU(nn.Module):
    r"""Applies the gated linear unit (GLU) function:

    .. math::

        \text{GLU}(x) = \text{Linear}_1(x) \otimes \sigma(\text{Linear}_2(x))


    where :math:`\otimes` is the element-wise multiplication operator and
    :math:`\sigma` is an activation function.

    Args:
        in_channels (int): Number of input features.
        hidden_channels (int, optional): Number of hidden features. Defaults to None, meaning hidden_channels=in_channels.
        activation (nn.Module, optional): Activation function to use. Defaults to Sigmoid.
    """

    def __init__(
        self, in_channels, hidden_channels=None, activation: Optional[nn.Module] = None
    ):
        super(GLU, self).__init__()
        self.act = nn.Sigmoid() if activation is None else activation
        hidden_channels = hidden_channels or in_channels
        self.W = nn.Linear(in_channels, hidden_channels)
        self.V = nn.Linear(in_channels, hidden_channels)

    def forward(self, x):
        return self.W(x) * self.act(self.V(x))


class ShiftedSoftplus(nn.Module):
    r"""Applies the ShiftedSoftplus function :math:`\text{ShiftedSoftplus}(x) = \frac{1}{\beta} *
    \log(1 + \exp(\beta * x))-\log(2)` element-wise.

    SoftPlus is a smooth approximation to the ReLU function and can be used
    to constrain the output of a machine to always be positive.
    """

    def __init__(self):
        super(ShiftedSoftplus, self).__init__()
        self.shift = torch.log(torch.tensor(2.0)).item()

    def forward(self, x):
        return F.softplus(x) - self.shift


class Swish(nn.Module):
    """Swish activation function as defined in https://arxiv.org/pdf/1710.05941 :

    .. math::

        \text{Swish}(x) = x \cdot \sigma(\beta x)

    Args:
        beta (float, optional): Scaling factor for Swish activation. Defaults to 1.

    """

    def __init__(self, beta=1.0):
        super(Swish, self).__init__()
        self.beta = beta

    def forward(self, x):
        return x * torch.sigmoid(self.beta * x)


class SwiGLU(nn.Module):
    """SwiGLU activation function as defined in https://arxiv.org/pdf/2002.05202 :

    .. math::

        \text{SwiGLU}(x) = \text{Linear}_1(x) \otimes \text{Swish}(\text{Linear}_2(x))

    W1, V have shape (in_features, hidden_features)
    Args:
        in_features (int): Number of input features.
        hidden_features (int, optional): Number of hidden features. Defaults to None, meaning hidden_features=in_features.
        beta (float, optional): Scaling factor for Swish activation. Defaults to 1.0.
    """

    def __init__(self, in_features, hidden_features=None, beta=1.0):
        super().__init__()
        hidden_features = hidden_features or in_features
        act = Swish(beta)
        self.glu = GLU(in_features, hidden_features, activation=act)

    def forward(self, x):
        return self.glu(x)


class CosineCutoff(nn.Module):
    def __init__(self, cutoff_lower=0.0, cutoff_upper=5.0):
        super(CosineCutoff, self).__init__()
        self.cutoff_lower = cutoff_lower
        self.cutoff_upper = cutoff_upper

    def forward(self, distances: Tensor) -> Tensor:
        if self.cutoff_lower > 0:
            cutoffs = 0.5 * (
                torch.cos(
                    math.pi
                    * (
                        2
                        * (distances - self.cutoff_lower)
                        / (self.cutoff_upper - self.cutoff_lower)
                        + 1.0
                    )
                )
                + 1.0
            )
            # remove contributions below the cutoff radius
            cutoffs = cutoffs * (distances < self.cutoff_upper)
            cutoffs = cutoffs * (distances > self.cutoff_lower)
            return cutoffs
        else:
            cutoffs = 0.5 * (torch.cos(distances * math.pi / self.cutoff_upper) + 1.0)
            # remove contributions beyond the cutoff radius
            cutoffs = cutoffs * (distances < self.cutoff_upper)
            return cutoffs


class MLP(nn.Module):
    r"""A simple multi-layer perceptron with a given number of layers and hidden channels.

    The simplest MLP has no hidden layers and is composed of two linear layers with a non-linear activation function in between:

    .. math::

        \text{MLP}(x) = \text{Linear}_o(\text{act}(\text{Linear}_i(x)))

    Where :math:`\text{Linear}_i` has input size :math:`\text{in_channels}` and output size :math:`\text{hidden_channels}` and :math:`\text{Linear}_o` has input size :math:`\text{hidden_channels}` and output size :math:`\text{out_channels}`.


    Args:
        in_channels (int): Number of input features.
        out_channels (int): Number of output features.
        hidden_channels (int): Number of hidden features.
        activation (str): Activation function to use.
        num_hidden_layers (int, optional): Number of hidden layers. Defaults to 0.
        dtype (torch.dtype, optional): Data type to use. Defaults to torch.float32.
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        hidden_channels,
        activation,
        num_hidden_layers=0,
        dtype=torch.float32,
    ):
        super(MLP, self).__init__()
        act_class = act_class_mapping[activation]
        self.act = act_class()
        self.layers = nn.Sequential()
        self.layers.append(nn.Linear(in_channels, hidden_channels, dtype=dtype))
        self.layers.append(self.act)
        for _ in range(num_hidden_layers):
            self.layers.append(nn.Linear(hidden_channels, hidden_channels, dtype=dtype))
            self.layers.append(self.act)
        self.layers.append(nn.Linear(hidden_channels, out_channels, dtype=dtype))

    def reset_parameters(self):
        for layer in self.layers:
            if isinstance(layer, nn.Linear):
                nn.init.xavier_uniform_(layer.weight)
                layer.bias.data.fill_(0)

    def forward(self, x):
        x = self.layers(x)
        return x


class GatedEquivariantBlock(nn.Module):
    """Gated Equivariant Block as defined in Schütt et al. (2021):
    Equivariant message passing for the prediction of tensorial properties and molecular spectra
    """

    def __init__(
        self,
        hidden_channels,
        out_channels,
        intermediate_channels=None,
        activation="silu",
        scalar_activation=False,
        dtype=torch.float,
    ):
        super(GatedEquivariantBlock, self).__init__()
        self.out_channels = out_channels

        if intermediate_channels is None:
            intermediate_channels = hidden_channels

        self.vec1_proj = nn.Linear(
            hidden_channels, hidden_channels, bias=False, dtype=dtype
        )
        self.vec2_proj = nn.Linear(
            hidden_channels, out_channels, bias=False, dtype=dtype
        )

        act_class = act_class_mapping[activation]
        self.update_net = MLP(
            in_channels=hidden_channels * 2,
            out_channels=out_channels * 2,
            hidden_channels=intermediate_channels,
            activation=activation,
            num_hidden_layers=0,
            dtype=dtype,
        )
        self.act = act_class() if scalar_activation else None

    def reset_parameters(self):
        nn.init.xavier_uniform_(self.vec1_proj.weight)
        nn.init.xavier_uniform_(self.vec2_proj.weight)
        self.update_net.reset_parameters()

    def forward(self, x, v):
        vec1_buffer = self.vec1_proj(v)

        # detach zero-entries to avoid NaN gradients during force loss backpropagation
        vec1 = torch.zeros(
            vec1_buffer.size(0),
            vec1_buffer.size(2),
            device=vec1_buffer.device,
            dtype=vec1_buffer.dtype,
        )
        mask = (vec1_buffer != 0).view(vec1_buffer.size(0), -1).any(dim=1)
        if not mask.all():
            warnings.warn(
                (
                    f"Skipping gradients for {(~mask).sum()} atoms due to vector features being zero. "
                    "This is likely due to atoms being outside the cutoff radius of any other atom. "
                    "These atoms will not interact with any other atom unless you change the cutoff."
                )
            )
        vec1[mask] = torch.norm(vec1_buffer[mask], dim=-2)

        vec2 = self.vec2_proj(v)

        x = torch.cat([x, vec1], dim=-1)
        x, v = torch.split(self.update_net(x), self.out_channels, dim=-1)
        v = v.unsqueeze(1) * vec2

        if self.act is not None:
            x = self.act(x)
        return x, v


def _broadcast(src: torch.Tensor, other: torch.Tensor, dim: int):
    """Broadcasts src to the shape of other along the given dimension."""
    if dim < 0:
        dim = other.dim() + dim
    if src.dim() == 1:
        for _ in range(0, dim):
            src = src.unsqueeze(0)
    for _ in range(src.dim(), other.dim()):
        src = src.unsqueeze(-1)
    src = src.expand(other.size())
    return src


def scatter(
    src: Tensor,
    index: Tensor,
    dim: int = 0,
    dim_size: Optional[int] = None,
    reduce: str = "sum",
) -> Tensor:
    """Has the signature of torch_scatter.scatter, but uses torch.scatter_reduce instead."""
    if dim_size is None:
        dim_size = index.max().item() + 1
    operation_dict = {
        "add": "sum",
        "sum": "sum",
        "mul": "prod",
        "mean": "mean",
        "min": "amin",
        "max": "amax",
    }
    reduce_op = operation_dict[reduce]
    # take into account the dimensionality of src
    index = _broadcast(index, src, dim)
    size = list(src.size())
    if dim_size is not None:
        size[dim] = dim_size
    elif index.numel() == 0:
        size[dim] = 0
    else:
        size[dim] = int(index.max()) + 1
    out = torch.zeros(size, dtype=src.dtype, device=src.device)
    res = out.scatter_reduce(dim, index, src, reduce_op)
    return res


rbf_class_mapping = {"gauss": GaussianSmearing, "expnorm": ExpNormalSmearing}

act_class_mapping = {
    "ssp": ShiftedSoftplus,
    "silu": nn.SiLU,
    "tanh": nn.Tanh,
    "sigmoid": nn.Sigmoid,
    "swish": Swish,
    "mish": nn.Mish,
}

dtype_mapping = {16: torch.float16, 32: torch.float, 64: torch.float64}



# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

import torch
from torch_geometric.transforms import Compose
from torch_geometric.datasets import QM9 as QM9_geometric
from torch_geometric.nn.models.schnet import qm9_target_dict
from torch_geometric.data import Data


class QM9(QM9_geometric):
    def __init__(self, root, transform=None, label=None):
        label2idx = dict(zip(qm9_target_dict.values(), qm9_target_dict.keys()))
        assert label in label2idx, (
            "Please pass the desired property to "
            'train on via "label". Available '
            f'properties are {", ".join(label2idx)}.'
        )

        self.label = label
        self.label_idx = label2idx[self.label]

        if transform is None:
            transform = self._filter_label
        else:
            transform = Compose([transform, self._filter_label])

        # Keep only pos, z and y in each sample
        def pre_transform(x):
            return Data(
                pos=x.pos,
                z=x.z,
                y=x.y,
            )

        super(QM9, self).__init__(
            root, transform=transform, pre_transform=pre_transform
        )

    def get_atomref(self, max_z=100):
        """Atomic energy reference values for the :py:mod:`torchmdnet.priors.Atomref` prior.

        Args:
            max_z (int): Maximum atomic number

        Returns:
            torch.Tensor: Atomic energy reference values for each element in the dataset.
        """
        atomref = self.atomref(self.label_idx)
        if atomref is None:
            return None
        if atomref.size(0) != max_z:
            tmp = torch.zeros(max_z).unsqueeze(1)
            idx = min(max_z, atomref.size(0))
            tmp[:idx] = atomref[:idx]
            return tmp
        return atomref

    def _filter_label(self, batch):
        batch.y = batch.y[:, self.label_idx].unsqueeze(1)
        return batch

    def download(self):
        super(QM9, self).download()

    def process(self):
        super(QM9, self).process()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Warp kernels for decomposing 3x3 tensors into I, A, S components."""

import warp as wp

from .utils import add_module, get_wp_fp_dtype


def generate_decompose_tensor(dtype: str, h_last: bool = True, use_irmem: bool = True):
    """Generate Warp kernels for decomposing a 3x3 tensor into I, A, S components."""
    dtype_wp = get_wp_fp_dtype(dtype)

    if not use_irmem:
        raise ValueError(f"only supporting use_irmem True, but got {use_irmem}")

    if not h_last:
        raise ValueError(f"only supporting h_last True but got {h_last}")

    class mat3x3(wp.types.matrix(shape=(3, 3), dtype=dtype_wp)):
        pass

    class vec3(wp.types.vector(length=3, dtype=dtype_wp)):
        pass

    class vec5(wp.types.vector(length=5, dtype=dtype_wp)):
        pass

    dim = 3 if use_irmem else 4

    def decompose_tensor_fwd(
        X: wp.array(ndim=4, dtype=dtype_wp),
        I: wp.array(ndim=dim, dtype=dtype_wp),
        A: wp.array(ndim=dim, dtype=dtype_wp),
        S: wp.array(ndim=dim, dtype=dtype_wp),
    ):
        b, h = wp.tid()

        X_reg = mat3x3()
        for i in range(3):
            for j in range(3):
                X_reg[i, j] = X[b, i, j, h]

        res = X.dtype(0)
        for i in range(3):
            res += X_reg[i, i]
        res = res / X.dtype(3.0)

        I[b, 0, h] = res

        denom = X.dtype(2.0)
        cnt = wp.int32(0)
        for i in range(2):
            for j in range(i + 1, 3):
                A[b, cnt, h] = (X_reg[i, j] - X_reg[j, i]) / denom
                cnt += 1

        cnt = wp.int32(0)
        for i in range(2):
            S[b, cnt, h] = X_reg[i, i] - res
            cnt += 1

            for j in range(i + 1, 3):
                S[b, cnt, h] = (X_reg[i, j] + X_reg[j, i]) / denom
                cnt += 1

    def decompose_tensor_bwd(
        dI: wp.array(ndim=dim, dtype=dtype_wp),
        dA: wp.array(ndim=dim, dtype=dtype_wp),
        dS: wp.array(ndim=dim, dtype=dtype_wp),
        dX: wp.array(ndim=4, dtype=dtype_wp),
    ):
        b, h = wp.tid()

        dX_reg = mat3x3(dX.dtype(0))

        dI_reg = dI[b, 0, h]
        dA_reg = vec3(dX.dtype(0))
        dS_reg = vec5(dX.dtype(0))

        for i in range(3):
            dA_reg[i] = dA[b, i, h]

        for i in range(5):
            dS_reg[i] = dS[b, i, h]

        for i in range(3):
            dX_reg[i, i] = dI_reg / dI.dtype(3.0)

        denom = dX.dtype(2.0)

        cnt = wp.int32(0)

        for i in range(3):
            for j in range(i + 1, 3):
                dX_reg[i, j] += dA_reg[cnt] / denom
                dX_reg[j, i] -= dA_reg[cnt] / denom
                cnt += 1

        cnt = wp.int32(0)
        for i in range(2):
            dX_reg[i, i] += dS_reg[cnt]
            for j in range(3):
                dX_reg[j, j] -= dS_reg[cnt] / dI.dtype(3.0)

            cnt += 1

            for j in range(i + 1, 3):
                dX_reg[i, j] += dS_reg[cnt] / denom
                dX_reg[j, i] += dS_reg[cnt] / denom
                cnt += 1

        for i in range(3):
            for j in range(3):
                dX[b, i, j, h] = dX_reg[i, j]

    def decompose_tensor_bwd_bwd(
        dX: wp.array(ndim=4, dtype=dtype_wp),
        d2I: wp.array(ndim=dim, dtype=dtype_wp),
        d2A: wp.array(ndim=dim, dtype=dtype_wp),
        d2S: wp.array(ndim=dim, dtype=dtype_wp),
    ):
        b, h = wp.tid()

        dX_reg = mat3x3(dX.dtype(0))
        d2I_reg = dX.dtype(0)
        d2A_reg = vec3(dX.dtype(0))
        d2S_reg = vec5(dX.dtype(0))

        for i in range(3):
            for j in range(3):
                dX_reg[i, j] = dX[b, i, j, h]

        for i in range(3):
            d2I_reg += dX_reg[i, i] / d2I.dtype(3.0)

        denom = dX.dtype(2.0)

        cnt = wp.int32(0)
        for i in range(3):
            for j in range(i + 1, 3):
                d2A_reg[cnt] += dX_reg[i, j] / denom
                d2A_reg[cnt] -= dX_reg[j, i] / denom
                cnt += 1

        cnt = wp.int32(0)
        for i in range(2):
            d2S_reg[cnt] += dX_reg[i, i]
            for j in range(3):
                d2S_reg[cnt] -= dX_reg[j, j] / d2I.dtype(3.0)
            cnt += 1

            for j in range(i + 1, 3):
                d2S_reg[cnt] += dX_reg[i, j] / denom
                d2S_reg[cnt] += dX_reg[j, i] / denom
                cnt += 1

        d2I[b, 0, h] = d2I_reg
        for i in range(3):
            d2A[b, i, h] = d2A_reg[i]

        for i in range(5):
            d2S[b, i, h] = d2S_reg[i]

    return (
        wp.Kernel(
            decompose_tensor_fwd,
            key=f"decompose_tensor_{dtype}",
            module=wp.get_module(f"decompose_tensor_{dtype}"),
        ),
        wp.Kernel(
            decompose_tensor_bwd,
            key=f"decompose_tensor_bwd_{dtype}",
            module=wp.get_module(f"decompose_tensor_bwd_{dtype}"),
        ),
        wp.Kernel(
            decompose_tensor_bwd_bwd,
            key=f"decompose_tensor_bwd_bwd_{dtype}",
            module=wp.get_module(f"decompose_tensor_bwd_bwd_{dtype}"),
        ),
    )


decompose_tensor_fwd_fp64, decompose_tensor_bwd_fp64, decompose_tensor_bwd_bwd_fp64 = generate_decompose_tensor(
    "float64"
)
decompose_tensor_fwd_fp32, decompose_tensor_bwd_fp32, decompose_tensor_bwd_bwd_fp32 = generate_decompose_tensor(
    "float32"
)
decompose_tensor_fwd_fp16, decompose_tensor_bwd_fp16, decompose_tensor_bwd_bwd_fp16 = generate_decompose_tensor(
    "float16"
)

add_module("decompose_tensor_fwd", ["float64"], decompose_tensor_fwd_fp64)
add_module("decompose_tensor_bwd", ["float64"], decompose_tensor_bwd_fp64)
add_module("decompose_tensor_bwd_bwd", ["float64"], decompose_tensor_bwd_bwd_fp64)

add_module("decompose_tensor_fwd", ["float32"], decompose_tensor_fwd_fp32)
add_module("decompose_tensor_bwd", ["float32"], decompose_tensor_bwd_fp32)
add_module("decompose_tensor_bwd_bwd", ["float32"], decompose_tensor_bwd_bwd_fp32)

add_module("decompose_tensor_fwd", ["float16"], decompose_tensor_fwd_fp16)
add_module("decompose_tensor_bwd", ["float16"], decompose_tensor_bwd_fp16)
add_module("decompose_tensor_bwd_bwd", ["float16"], decompose_tensor_bwd_bwd_fp16)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Warp kernels for O(3)-equivariant 3x3 tensor matrix multiplication."""

import warp as wp

from .utils import add_module, get_wp_fp_dtype


def generate_tensor_matmul_o3_3x3(dtype: str):
    """Generate Warp kernels for O(3)-equivariant 3x3 matrix multiplication: C = AB + BA."""
    dtype_wp = get_wp_fp_dtype(dtype)

    class mat3x3(wp.types.matrix(shape=(3, 3), dtype=dtype_wp)):
        pass

    def tensor_matmul_o3_3x3_fwd(
        A: wp.array(ndim=4, dtype=dtype_wp),
        B: wp.array(ndim=4, dtype=dtype_wp),
        C: wp.array(ndim=4, dtype=dtype_wp),
    ):
        b, h = wp.tid()

        a_reg = mat3x3()
        b_reg = mat3x3()
        c_reg = mat3x3()

        for i in range(3):
            for j in range(3):
                a_reg[i, j] = A[b, i, j, h]
                b_reg[i, j] = B[b, i, j, h]

        for i in range(3):
            for j in range(3):
                for k in range(3):
                    c_reg[i, j] += a_reg[i, k] * b_reg[k, j] + b_reg[i, k] * a_reg[k, j]

        for i in range(3):
            for j in range(3):
                C[b, i, j, h] = c_reg[i, j]

    def tensor_matmul_o3_3x3_bwd(
        A: wp.array(ndim=4, dtype=dtype_wp),
        B: wp.array(ndim=4, dtype=dtype_wp),
        dC: wp.array(ndim=4, dtype=dtype_wp),
        dA: wp.array(ndim=4, dtype=dtype_wp),
        dB: wp.array(ndim=4, dtype=dtype_wp),
    ):
        b, h = wp.tid()

        a_reg = mat3x3()
        b_reg = mat3x3()

        da_reg = mat3x3()
        db_reg = mat3x3()

        dc_reg = mat3x3()

        for i in range(3):
            for j in range(3):
                a_reg[i, j] = A[b, i, j, h]
                b_reg[i, j] = B[b, i, j, h]
                dc_reg[i, j] = dC[b, i, j, h]

        for i in range(3):
            for j in range(3):
                for k in range(3):
                    da_reg[i, j] += dc_reg[i, k] * b_reg[j, k]
                    da_reg[j, k] += dc_reg[i, k] * b_reg[i, j]
                    db_reg[i, j] += dc_reg[i, k] * a_reg[j, k]
                    db_reg[j, k] += dc_reg[i, k] * a_reg[i, j]

        for i in range(3):
            for j in range(3):
                dA[b, i, j, h] = da_reg[i, j]
                dB[b, i, j, h] = db_reg[i, j]

    def tensor_matmul_o3_3x3_bwd_bwd(
        A: wp.array(ndim=4, dtype=dtype_wp),
        B: wp.array(ndim=4, dtype=dtype_wp),
        dA: wp.array(ndim=4, dtype=dtype_wp),
        dB: wp.array(ndim=4, dtype=dtype_wp),
        dC: wp.array(ndim=4, dtype=dtype_wp),
        d2A: wp.array(ndim=4, dtype=dtype_wp),
        d2B: wp.array(ndim=4, dtype=dtype_wp),
        d2C: wp.array(ndim=4, dtype=dtype_wp),
    ):
        b, h = wp.tid()

        a_reg = mat3x3()
        b_reg = mat3x3()

        da_reg = mat3x3()
        db_reg = mat3x3()

        dc_reg = mat3x3()

        d2a_reg = mat3x3()
        d2b_reg = mat3x3()

        d2c_reg = mat3x3()

        for i in range(3):
            for j in range(3):
                a_reg[i, j] = A[b, i, j, h]
                b_reg[i, j] = B[b, i, j, h]

                da_reg[i, j] = dA[b, i, j, h]
                db_reg[i, j] = dB[b, i, j, h]

                dc_reg[i, j] = dC[b, i, j, h]

        for i in range(3):
            for j in range(3):
                for k in range(3):
                    d2a_reg[i, j] += dc_reg[i, k] * db_reg[j, k]
                    d2a_reg[j, i] += dc_reg[k, i] * db_reg[k, j]

                    d2b_reg[i, j] += dc_reg[i, k] * da_reg[j, k]
                    d2b_reg[j, i] += dc_reg[k, i] * da_reg[k, j]

        for i in range(3):
            for j in range(3):
                for k in range(3):
                    #  grad_grad_x @ y + x @ grad_grad_y
                    d2c_reg[i, j] += da_reg[i, k] * b_reg[k, j]
                    d2c_reg[i, j] += a_reg[i, k] * db_reg[k, j]

                    d2c_reg[i, j] += db_reg[i, k] * a_reg[k, j]
                    d2c_reg[i, j] += b_reg[i, k] * da_reg[k, j]

        for i in range(3):
            for j in range(3):
                d2A[b, i, j, h] = d2a_reg[i, j]
                d2B[b, i, j, h] = d2b_reg[i, j]
                d2C[b, i, j, h] = d2c_reg[i, j]

    return (
        wp.Kernel(
            tensor_matmul_o3_3x3_fwd,
            key=f"tensor_matmul_o3_3x3_{dtype}",
            module=wp.get_module(f"tensor_matmul_o3_3x3_{dtype}"),
        ),
        wp.Kernel(
            tensor_matmul_o3_3x3_bwd,
            key=f"tensor_matmul_o3_3x3_bwd_{dtype}",
            module=wp.get_module(f"tensor_matmul_o3_3x3_bwd_{dtype}"),
        ),
        wp.Kernel(
            tensor_matmul_o3_3x3_bwd_bwd,
            key=f"tensor_matmul_o3_3x3_bwd_bwd_{dtype}",
            module=wp.get_module(f"tensor_matmul_o3_3x3_bwd_bwd_{dtype}"),
        ),
    )


(
    tensor_matmul_o3_3x3_fwd_fp64,
    tensor_matmul_o3_3x3_bwd_fp64,
    tensor_matmul_o3_3x3_bwd_bwd_fp64,
) = generate_tensor_matmul_o3_3x3("float64")
(
    tensor_matmul_o3_3x3_fwd_fp32,
    tensor_matmul_o3_3x3_bwd_fp32,
    tensor_matmul_o3_3x3_bwd_bwd_fp32,
) = generate_tensor_matmul_o3_3x3("float32")
(
    tensor_matmul_o3_3x3_fwd_fp16,
    tensor_matmul_o3_3x3_bwd_fp16,
    tensor_matmul_o3_3x3_bwd_bwd_fp16,
) = generate_tensor_matmul_o3_3x3("float16")

add_module("tensor_matmul_o3_3x3_fwd", ["float64"], tensor_matmul_o3_3x3_fwd_fp64)
add_module("tensor_matmul_o3_3x3_bwd", ["float64"], tensor_matmul_o3_3x3_bwd_fp64)
add_module(
    "tensor_matmul_o3_3x3_bwd_bwd", ["float64"], tensor_matmul_o3_3x3_bwd_bwd_fp64
)

add_module("tensor_matmul_o3_3x3_fwd", ["float32"], tensor_matmul_o3_3x3_fwd_fp32)
add_module("tensor_matmul_o3_3x3_bwd", ["float32"], tensor_matmul_o3_3x3_bwd_fp32)
add_module(
    "tensor_matmul_o3_3x3_bwd_bwd", ["float32"], tensor_matmul_o3_3x3_bwd_bwd_fp32
)

add_module("tensor_matmul_o3_3x3_fwd", ["float16"], tensor_matmul_o3_3x3_fwd_fp16)
add_module("tensor_matmul_o3_3x3_bwd", ["float16"], tensor_matmul_o3_3x3_bwd_fp16)
add_module(
    "tensor_matmul_o3_3x3_bwd_bwd", ["float16"], tensor_matmul_o3_3x3_bwd_bwd_fp16
)



# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

from torch_geometric.data import Data, Dataset
import numpy as np
import torch as pt
import os


class MemmappedDataset(Dataset):
    """Dataset class which stores all the molecular data in memory-mapped files.

    It supports the following attributes in the data returned by sample_iter:

        - :obj:`z`: Atomic numbers of the atoms.
        - :obj:`pos`: Positions of the atoms.
        - :obj:`y`: Energy of the conformation.
        - :obj:`neg_dy`: Forces on the atoms.
        - :obj:`q`: Total charge of the conformation.
        - :obj:`pq`: Partial charges of the atoms.
        - :obj:`dp`: Dipole moment of the conformation.

    The data is stored in the following files:

        - :obj:`name.idx.mmap`: Index of the first atom of each conformation.
        - :obj:`name.z.mmap`: Atomic numbers of all the atoms.
        - :obj:`name.pos.mmap`: Positions of all the atoms.
        - :obj:`name.y.mmap`: Energy of each conformation.
        - :obj:`name.neg_dy.mmap`: Forces on all the atoms.
        - :obj:`name.q.mmap`: Total charge of each conformation.
        - :obj:`name.pq.mmap`: Partial charges of all the atoms.
        - :obj:`name.dp.mmap`: Dipole moment of each conformation.

    Args:
        root (str): Root directory where the dataset should be stored.
        transform (callable, optional): A function/transform that takes in an
            :obj:`torch_geometric.data.Data` object and returns a transformed
            version. The data object will be transformed before every access.
        pre_transform (callable, optional): A function/transform that takes in an
            :obj:`torch_geometric.data.Data` object and returns a transformed
            version. The data object will be transformed before being saved to disk.
        pre_filter (callable, optional): A function that takes in an
            :obj:`torch_geometric.data.Data` object and returns a boolean value,
            indicating whether the data object should be included in the final
            dataset.
        properties (tuple of str, optional): The properties to include in the
            dataset. Can be any subset of :obj:`y`, :obj:`neg_dy`, :obj:`q`,
            :obj:`pq`, and :obj:`dp`.
    """

    def __init__(
        self,
        root,
        transform=None,
        pre_transform=None,
        pre_filter=None,
        properties=("y", "neg_dy", "q", "pq", "dp"),
    ):
        if not hasattr(self, "name"):
            self.name = self.__class__.__name__
        self.properties = properties
        super().__init__(root, transform, pre_transform, pre_filter)

        fnames = self.processed_paths_dict

        self.mmaps = {}
        self.mmaps["idx"] = np.memmap(fnames["idx"], mode="r", dtype=np.int64)
        self.mmaps["z"] = np.memmap(fnames["z"], mode="r", dtype=np.int8)
        num_all_confs = self.mmaps["idx"].shape[0] - 1
        num_all_atoms = self.mmaps["z"].shape[0]
        self.mmaps["pos"] = np.memmap(
            fnames["pos"], mode="r", dtype=np.float32, shape=(num_all_atoms, 3)
        )
        if "y" in self.properties:
            self.mmaps["y"] = np.memmap(fnames["y"], mode="r", dtype=np.float64)
        if "neg_dy" in self.properties:
            self.mmaps["neg_dy"] = np.memmap(
                fnames["neg_dy"], mode="r", dtype=np.float32, shape=(num_all_atoms, 3)
            )
        if "q" in self.properties:
            self.mmaps["q"] = np.memmap(fnames["q"], mode="r", dtype=np.int8)
        if "pq" in self.properties:
            self.mmaps["pq"] = np.memmap(fnames["pq"], mode="r", dtype=np.float32)
        if "dp" in self.properties:
            self.mmaps["dp"] = np.memmap(
                fnames["dp"], mode="r", dtype=np.float32, shape=(num_all_confs, 3)
            )

        assert self.mmaps["idx"][0] == 0
        assert self.mmaps["idx"][-1] == len(self.mmaps["z"])
        assert len(self.mmaps["idx"]) == len(self.mmaps["y"]) + 1

    @property
    def processed_file_names(self):
        return [
            f"{self.name}.{prop}.mmap"
            for prop in ["idx", "z", "pos"] + list(self.properties)
        ]

    @property
    def processed_paths_dict(self):
        return {
            prop: fname
            for prop, fname in zip(
                ["idx", "z", "pos"] + list(self.properties), self.processed_paths
            )
        }

    def sample_iter(self, mol_ids=False):
        raise NotImplementedError()

    def process(self):
        import gc

        print("Gathering statistics...")
        num_all_confs = 0
        num_all_atoms = 0
        for data in self.sample_iter():
            num_all_confs += 1
            num_all_atoms += data.z.shape[0]

        print(f"  Total number of conformers: {num_all_confs}")
        print(f"  Total number of atoms: {num_all_atoms}")
        print(f"  Properties available: {self.properties}")

        fnames = self.processed_paths_dict

        mmaps = {}
        mmaps["idx"] = np.memmap(
            fnames["idx"] + ".tmp",
            mode="w+",
            dtype=np.int64,
            shape=(num_all_confs + 1,),
        )
        mmaps["z"] = np.memmap(
            fnames["z"] + ".tmp", mode="w+", dtype=np.int8, shape=(num_all_atoms,)
        )
        mmaps["pos"] = np.memmap(
            fnames["pos"] + ".tmp",
            mode="w+",
            dtype=np.float32,
            shape=(num_all_atoms, 3),
        )
        if "y" in self.properties:
            mmaps["y"] = np.memmap(
                fnames["y"] + ".tmp",
                mode="w+",
                dtype=np.float64,
                shape=(num_all_confs,),
            )
        if "neg_dy" in self.properties:
            mmaps["neg_dy"] = np.memmap(
                fnames["neg_dy"] + ".tmp",
                mode="w+",
                dtype=np.float32,
                shape=(num_all_atoms, 3),
            )
        if "q" in self.properties:
            mmaps["q"] = np.memmap(
                fnames["q"] + ".tmp", mode="w+", dtype=np.int8, shape=num_all_confs
            )
        if "pq" in self.properties:
            mmaps["pq"] = np.memmap(
                fnames["pq"] + ".tmp",
                mode="w+",
                dtype=np.float32,
                shape=num_all_atoms,
            )
        if "dp" in self.properties:
            mmaps["dp"] = np.memmap(
                fnames["dp"] + ".tmp",
                mode="w+",
                dtype=np.float32,
                shape=(num_all_confs, 3),
            )

        print("Storing data...")
        i_atom = 0
        for i_conf, data in enumerate(self.sample_iter()):
            i_next_atom = i_atom + data.z.shape[0]

            mmaps["idx"][i_conf] = i_atom
            mmaps["z"][i_atom:i_next_atom] = data.z.to(pt.int8)
            mmaps["pos"][i_atom:i_next_atom] = data.pos
            if "y" in self.properties:
                mmaps["y"][i_conf] = data.y
            if "neg_dy" in self.properties:
                mmaps["neg_dy"][i_atom:i_next_atom] = data.neg_dy
            if "q" in self.properties:
                mmaps["q"][i_conf] = data.q.to(pt.int8)
            if "pq" in self.properties:
                mmaps["pq"][i_atom:i_next_atom] = data.pq
            if "dp" in self.properties:
                mmaps["dp"][i_conf] = data.dp
            i_atom = i_next_atom

        mmaps["idx"][-1] = num_all_atoms
        assert i_atom == num_all_atoms

        for prop in list(mmaps.keys()):
            mmaps[prop].flush()
            del mmaps[prop]

        # Force garbage collection to ensure files are closed before renaming
        gc.collect()

        os.rename(fnames["idx"] + ".tmp", fnames["idx"])
        os.rename(fnames["z"] + ".tmp", fnames["z"])
        os.rename(fnames["pos"] + ".tmp", fnames["pos"])
        if "y" in self.properties:
            os.rename(fnames["y"] + ".tmp", fnames["y"])
        if "neg_dy" in self.properties:
            os.rename(fnames["neg_dy"] + ".tmp", fnames["neg_dy"])
        if "q" in self.properties:
            os.rename(fnames["q"] + ".tmp", fnames["q"])
        if "pq" in self.properties:
            os.rename(fnames["pq"] + ".tmp", fnames["pq"])
        if "dp" in self.properties:
            os.rename(fnames["dp"] + ".tmp", fnames["dp"])

    def len(self):
        return len(self.mmaps["idx"]) - 1

    def get(self, idx):
        """Gets the data object at index :obj:`idx`.

        The data object contains the following attributes:

            - :obj:`z`: Atomic numbers of the atoms.
            - :obj:`pos`: Positions of the atoms.
            - :obj:`y`: Formation energy of the molecule.
            - :obj:`neg_dy`: Forces on the atoms.
            - :obj:`q`: Total charge of the molecule.
            - :obj:`pq`: Partial charges of the atoms.
            - :obj:`dp`: Dipole moment of the molecule.

        Args:
            idx (int): Index of the data object.

        Returns:
            :obj:`torch_geometric.data.Data`: The data object.
        """
        atoms = slice(self.mmaps["idx"][idx], self.mmaps["idx"][idx + 1])
        z = pt.tensor(self.mmaps["z"][atoms], dtype=pt.long)
        pos = pt.tensor(self.mmaps["pos"][atoms])

        props = {}
        if "y" in self.properties:
            props["y"] = pt.tensor(self.mmaps["y"][idx]).view(1, 1)
        if "neg_dy" in self.properties:
            props["neg_dy"] = pt.tensor(self.mmaps["neg_dy"][atoms])
        if "q" in self.properties:
            props["q"] = pt.tensor(self.mmaps["q"][idx], dtype=pt.long)
        if "pq" in self.properties:
            props["pq"] = pt.tensor(self.mmaps["pq"][atoms])
        if "dp" in self.properties:
            props["dp"] = pt.tensor(self.mmaps["dp"][idx])
        return Data(z=z, pos=pos, **props)

    def __del__(self):
        # Flush and close all the memory-mapped files
        import gc

        for prop in list(self.mmaps.keys()):
            self.mmaps[prop].flush()
            del self.mmaps[prop]

        gc.collect()
